from django.urls import reverse
from django.contrib import admin

from munch.core.utils import pretty_json_as_html

from .models import Mail
from .models import MailBatch
from .models import MailStatus


class MailStatusInline(admin.TabularInline):
    extra = 0
    max_num = 0
    can_delete = False
    model = MailStatus
    ordering = ['creation_date']
    readonly_fields = [
        'source_ip', 'destination_domain', 'status_code',
        'source_hostname', 'raw_msg', 'creation_date', 'status']


class MailInline(admin.TabularInline):
    extra = 0
    max_num = 0
    model = Mail
    fields = ['sender', 'recipient', 'author_url', 'message_field']
    readonly_fields = [
        'identifier', 'message', 'sender', 'recipient',
        'author_url', 'message_field']
    exclude = ['author']
    ordering = ['creation_date']

    def author_url(self, instance):
        return '<a target="_blank" href="{}">{}</a>'.format(
            reverse('admin:users_munchuser_change', args=[instance.author.pk]),
            instance.author)
    author_url.allow_tags = True
    author_url.short_description = 'Author'

    def message_field(self, instance):
        if instance.message:
            return 'Still have a message associated (pk:{})'.format(
                instance.message.pk)
        else:
            return 'No message associated'
    message_field.short_description = 'Message'


class MailBatchAdmin(admin.ModelAdmin):
    inlines = [MailInline]
    search_fields = ['id', 'identifier', 'name']
    list_display = [
        'identifier', 'name', 'creation_date', 'author_url', 'category_url']
    list_filter = ['creation_date']

    def get_readonly_fields(self, request, obj=None):
        if obj:
            return self.readonly_fields + (
                'pretty_msg_links', 'author', '_links')
        return self.readonly_fields

    def _links(self, instance):
        content = '<ul>'
        content += '<li>Category: {}</li>'.format(self.category_url(instance))
        content += '<li>Author: {}</li>'.format(self.author_url(instance))
        content += '</ul>'
        return content
    _links.allow_tags = True
    _links.short_description = 'Links'

    def pretty_msg_links(self, instance):
        return pretty_json_as_html(instance.msg_links)
    pretty_msg_links.short_description = 'msg_links'

    def author_url(self, instance):
        return '<a target="_blank" href="{}">{}</a>'.format(
            reverse('admin:users_munchuser_change', args=[instance.author.pk]),
            instance.author)
    author_url.allow_tags = True
    author_url.short_description = 'Author'

    def category_url(self, instance):
        if instance.category:
            return '<a target="_blank" href="{}">{}</a>'.format(
                reverse(
                    'admin:core_category_change', args=[instance.category.pk]),
                instance.category)
        else:
            return '-'
    category_url.allow_tags = True
    category_url.short_description = 'Category'


class MailAdmin(admin.ModelAdmin):
    inlines = [MailStatusInline]
    search_fields = [
        'id', 'identifier', 'headers', 'sender', 'recipient']
    list_display = [
        'identifier', 'sender', 'recipient', 'creation_date', 'curstatus',
        'track_clicks', 'track_open', 'batch_url', 'author_url']
    list_filter = ['creation_date', 'curstatus', 'track_clicks', 'track_open']

    def get_readonly_fields(self, request, obj=None):
        if obj:
            return self.readonly_fields + (
                'pretty_headers', 'pretty_msg_links', 'identifier',
                'track_open', 'track_clicks', 'first_status_date', 'batch',
                'author', 'latest_status_date', 'message', 'creation_date',
                'had_delay', 'sender', 'delivery_duration', 'recipient',
                'curstatus', 'message_field', '_links')
        return self.readonly_fields

    def _links(self, instance):
        content = '<ul>'
        content += '<li>Batch: {}</li>'.format(self.batch_url(instance))
        content += '<li>Author: {}</li>'.format(self.author_url(instance))
        content += '</ul>'
        return content
    _links.allow_tags = True
    _links.short_description = 'Links'

    def author_url(self, instance):
        return '<a target="_blank" href="{}">{}</a>'.format(
            reverse('admin:users_munchuser_change', args=[instance.author.pk]),
            instance.author)
    author_url.allow_tags = True
    author_url.short_description = 'Author'

    def batch_url(self, instance):
        if instance.batch:
            return '<a target="_blank" href="{}">{}</a>'.format(
                reverse(
                    'admin:transactional_mailbatch_change',
                    args=[instance.batch.pk]),
                instance.batch)
        else:
            return 'No batch'
    batch_url.allow_tags = True
    batch_url.short_description = 'Batch'

    def message_field(self, instance):
        if instance.message:
            return 'Still have a message associated (pk:{})'.format(
                instance.message.pk)
        else:
            return 'No message associated'
    message_field.short_description = 'Message'

    def pretty_headers(self, instance):
        return pretty_json_as_html(instance.headers)
    pretty_headers.short_description = 'headers'

    def pretty_msg_links(self, instance):
        return pretty_json_as_html(instance.msg_links)
    pretty_msg_links.short_description = 'msg_links'


admin.site.register(Mail, MailAdmin)
admin.site.register(MailBatch, MailBatchAdmin)



import logging
from datetime import datetime
from unittest.mock import patch
from unittest.mock import MagicMock

import celery
import requests
from django.test import TestCase
from django.conf import settings
from libfaketime import fake_time
from slimta.smtp.reply import Reply
from slimta.envelope import Envelope

from munch.core.mail.utils import mk_base64_uuid

from ..status import send_webhook
from ..status import handle_dsn_status
from ..status import handle_smtp_status

from .factories import MailFactory


def mock_status(status_code):
    m = MagicMock()
    m.status_code = status_code
    return m


class TestDSNStatus(TestCase):
    def setUp(self):
        identifier = mk_base64_uuid()
        self.mail = MailFactory(
            identifier=identifier,
            headers={
                'To': 'someone@example.com',
                'Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)',
                settings.TRANSACTIONAL['X_MESSAGE_ID_HEADER']: identifier,
                settings.TRANSACTIONAL['X_HTTP_DSN_RETURN_PATH_HEADER']: 'http://example.com/ping',  # noqa
            })

    def test_push_dsn_status(self):
        with patch('requests.post', return_value=mock_status(200)):
            handle_dsn_status(
                (
                    'To: return-{}@test.munch.example.com\n'
                    'Date: Fri,  5 Aug 2014 23:35:50 +0700 (WIT)').format(
                        self.mail.identifier),
                {
                    'Diagnostic-Code': 'smtp; 200 Delivered',
                    'Status': '2.0.0',
                    'Arrival-Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)',
                    'Final-Recipient': 'foo@bar'
                }
            )
            requests.post.assert_called_with('http://example.com/ping', json={
                'status': 'delivered',
                'message': 'smtp; 200 Delivered',
                'esmtp_status': '2.0.0',
                'smtp_status': '200',
                'date': '2014-08-05T16:35:50+00:00',
                'recipient': 'foo@bar'
            })

    def test_push_dsn_status_no_smtp_diagnostic(self):
        with patch('requests.post', return_value=mock_status(200)):
            handle_dsn_status(
                (
                    'To: return-{}@test.munch.example.com\n'
                    'Date: Fri,  5 Aug 2014 23:35:50 +0700 (WIT)').format(
                        self.mail.identifier),
                {
                    'Diagnostic-Code': 'FreeForm!',
                    'Status': '2.0.0',
                    'Arrival-Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)',
                    'Final-Recipient': 'foo@bar'
                }
            )
            requests.post.assert_called_with('http://example.com/ping', json={
                'status': 'delivered',
                'message': 'FreeForm!',
                'esmtp_status': '2.0.0',
                'smtp_status': 'unknown',
                'date': '2014-08-05T16:35:50+00:00',
                'recipient': 'foo@bar'
            })

    def test_push_dsn_status_no_arrival_date_nor_date(self):
        with fake_time('2015-12-18'):
            with patch('requests.post', return_value=mock_status(200)):
                handle_dsn_status(
                    'To: return-{}@test.munch.example.com'.format(
                        self.mail.identifier),
                    {
                        'Diagnostic-Code': 'FreeForm!',
                        'Status': '2.0.0',
                        'Final-Recipient': 'foo@bar'
                    }
                )
                requests.post.assert_called_with(
                    'http://example.com/ping',
                    json={
                        'status': 'delivered',
                        'message': 'FreeForm!',
                        'esmtp_status': '2.0.0',
                        'smtp_status': 'unknown',
                        'date': '2015-12-18T00:00:00',
                        'recipient': 'foo@bar'
                    })

    def test_push_dsn_status_no_arrival_date(self):
        with fake_time('2015-12-18'):
            with patch('requests.post', return_value=mock_status(200)):
                handle_dsn_status(
                    (
                        'To: return-{}@test.munch.example.com\n'
                        'Date: Fri,  5 Aug 2014 23:35:50 +0700 (WIT)').format(
                            self.mail.identifier),
                    {
                        'Diagnostic-Code': 'FreeForm!',
                        'Status': '2.0.0',
                        'Final-Recipient': 'foo@bar'
                    }
                )
                requests.post.assert_called_with(
                    'http://example.com/ping',
                    json={
                        'status': 'delivered',
                        'message': 'FreeForm!',
                        'esmtp_status': '2.0.0',
                        'smtp_status': 'unknown',
                        'date': '2014-08-05T16:35:50+00:00',
                        'recipient': 'foo@bar'
                    }
                )

    def test_push_dsn_status_webhook_error(self):
        with self.assertRaises(celery.exceptions.Retry):
            with patch('requests.post', return_value=mock_status(500)):
                send_webhook(
                    self.mail.identifier,
                    {
                        'To': 'return-{}@test.munch.example.com'.format(
                            self.mail.identifier),
                        'Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)',
                        'Diagnostic-Code': 'smtp; 200 Delivered',
                        'Status': '2.0.0',
                        'Arrival-Date': 'Fri, 5 Aug 2014 23:35:50 +0700 (WIT)',
                    },
                    {
                        'status': 'Delivered',
                        'message': '200 Delivered',
                        'smtp_status': 'delivered',
                        'esmtp_status': '2.0.0',
                        'date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)',
                        'recipient': ''
                    })

    def test_push_dsn_status_broken_headers(self):
        with self.assertRaises(celery.exceptions.Reject):
            with patch('requests.post', return_value=mock_status(201)):
                handle_dsn_status(
                    'Date: Fri,  5 Aug 2014 23:35:50 +0700 (WIT)',
                    {
                        'Diagnostic-Code': 'smtp; 200 Delivered',
                        'Status': '2.0.0',
                        'Arrival-Date': 'Fri, 5 Aug 2014 23:35:50 +0700 (WIT)',
                    }
                )
                requests.post.assert_called_with(
                    'http://example.com/ping',
                    json={
                        'status': 'delivered',
                        'message': 'smtp; 200 Delivered',
                        'smtp_status': '2.0.0',
                        'date': '2014-08-05T16:35:50+00:00',
                        'recipient': ''
                    })

    def test_push_dsn_status_missing_esmtp_code(self):
        with self.assertRaises(celery.exceptions.Reject):
            with patch('requests.post', return_value=mock_status(201)):
                handle_dsn_status(
                    (
                        'To: return-{}@test.munch.example.com\n'
                        'Date: Fri,  5 Aug 2014 23:35:50 +0700 (WIT)').format(
                            self.mail.identifier),
                    {
                        'Diagnostic-Code': 'smtp; 200 Delivered'
                    }
                )

    def test_push_dsn_status_unknown_email(self):
        with self.assertRaises(celery.exceptions.Reject):
            with patch('requests.post', return_value=mock_status(201)):
                # This status is unknown to the DB
                handle_dsn_status(
                    (
                        'To: return-4242424242424242424242@'
                        'test.munch.example.com\n'
                        'Date: Fri,  5 Aug 2014 23:35:50 +0700 (WIT)'),
                    {
                        'Diagnostic-Code': 'smtp; 200 Delivered'
                    }
                )

    def test_push_smtp_status(self):
        body = (
            "Return-Path: return-{}@test.munch.example.com\n"
            "X-CM-Message-Id: {}".format(
                self.mail.identifier, self.mail.identifier))
        env = Envelope()
        env.parse(body.encode('utf-8'))
        reply = Reply('200', '2.0.0 Delivered')

        with patch('requests.post', return_value=mock_status(201)):
            with fake_time('2015-12-18'):
                handle_smtp_status(
                    'delivered', datetime.now(),
                    '{}'.format(self.mail.identifier), reply, 'localhost')

            requests.post.assert_called_with(
                'http://example.com/ping',
                json={
                    'status': 'delivered',
                    'message': '2.0.0 Delivered',
                    'smtp_status': '200',
                    'esmtp_status': '2.0.0',
                    'date': '2015-12-18T00:00:00',
                    'recipient': ''
                })

    def test_push_smtp_status_no_http_returnpath(self):
        mail = MailFactory()
        env = Envelope()
        env.parse("Return-Path: return-{}@test.munch.example.com".format(
            mail.identifier).encode('utf-8'))
        reply = Reply('200', '2.0.0 Delivered')

        with patch('requests.post'):
            with patch('logging.warning'):
                handle_smtp_status(
                    'delivered', datetime.now(),
                    mail.identifier, reply, 'localhost')

                requests.post.assert_not_called()
                logging.warning.assert_called()



import logging

from django.core.signals import request_started
from django.core.signals import request_finished
from slimta.edge.smtp import SmtpEdge
from slimta.edge.smtp import SmtpValidators
from slimta.util.proxyproto import ProxyProtocol
from slimta.util.proxyproto import ProxyProtocolV1
from slimta.util.proxyproto import ProxyProtocolV2
from slimta.util.proxyproto import LocalConnection
from slimta.util.proxyproto import invalid_pp_source_address

from munch.apps.users.models import SmtpApplication

log = logging.getLogger(__name__)


class EdgeValidators(SmtpValidators):
    def handle_auth(self, reply, creds):
        # Force authentication over TLS
        if not self.session.security:
            reply.code = '530'
            reply.message = '5.7.0 Must issue a STARTTLS command first'
            return

        authenticated = False
        application = SmtpApplication.objects.filter(
            username=creds.authcid, author__is_active=True).only(
                'secret').first()
        if application:
            authenticated = creds.check_secret(application.secret)
            if authenticated:
                log.info('Successfull login from "{}"'.format(creds.authcid))

        if not authenticated:
            reply.code = '535'
            reply.message = '5.7.8 Authentication credentials invalid'
            log.warning('Failed login from "{}" (wrong credentials)'.format(
                creds.authcid))

    def handle_mail(self, reply, sender, params):
        # Make authentication mandatory
        if not hasattr(self.session, 'auth') or not self.session.auth:
            reply.code = '535'
            reply.message = '5.7.0 Authentification is required'
            log.warning('Unauthenticated traffic from {} refused'.format(
                self.session.address[0]))


class TransactionalSmtpEdge(SmtpEdge):
    """ Default Transactional Smtp Edge """
    def handle(self, socket, address):
        """ Just overwrite the handle to have database lock per connection """
        log.info('Incoming connection from {}:{}'.format(
            address[0], address[1]))
        request_started.send(
            sender='transactional-edge-{}-{}'.format(address[0], address[1]))
        super().handle(socket, address)
        request_finished.send(
            sender='transactional-edge-{}-{}'.format(address[0], address[1]))


class ProxyProtocolTransactionalSmtpEdge(
        ProxyProtocol, TransactionalSmtpEdge):
    """ Proxy protocol compatible Smtp Edge based on TransactionalSmtpEdge """
    def handle(self, socket, address):
        """
        This method is mostly copy-paste from Slimta to be able to handle
        remote ip retrieved before calling TransactionalSmtpEdge.handle method
        """
        try:
            initial = self._ProxyProtocol__read_pp_initial(socket)
            if initial.startswith(b'PROXY '):
                address, _ = ProxyProtocolV1.process_pp_v1(socket, initial)
            elif initial == b'\r\n\r\n\x00\r\nQ':
                address, _ = ProxyProtocolV2.process_pp_v2(socket, initial)
            else:
                raise AssertionError('Invalid proxy protocol signature')
        except LocalConnection:
            log.error(
                "Parsed proxy protocol header si for local "
                "connection, and should not be proxied.")
            return
        except AssertionError as exc:
            log.error(
                'Error while retrieving ProxyProto '
                'address: {} ({})'.format(socket, str(exc)))
            address = invalid_pp_source_address
        else:
            log.debug(
                'Successfully retrieve ProxyProto address: {}:{}'.format(
                    address[0], address[1]))
        super(TransactionalSmtpEdge, self).handle(socket, address)



from django.http import Http404
from django.conf import settings
from django.utils import timezone
from django.shortcuts import render
from django.db.utils import IntegrityError
from django.core.exceptions import PermissionDenied
from django.core.exceptions import FieldDoesNotExist
from django.views.decorators.csrf import csrf_exempt
from rest_framework.serializers import ValidationError
from ipware.ip import get_ip

from .models import Contact
from .models import ContactList
from .models import ContactQueue
from .models import AbstractContact
from .models import CollectedContact
from .forms import ContactSubscriptionForm
from .forms import CollectedContactSubscriptionForm
from .api.v1.serializers import PropertiesFieldHelper


def test_form(request, uuid):
    contact_list = None
    try:
        contact_list = ContactQueue.objects.get(uuid=uuid)
        contact_model = CollectedContact
        subscription_form = CollectedContactSubscriptionForm
    except ContactQueue.DoesNotExist:
        contact_list = ContactList.objects.get(uuid=uuid)
        contact_model = Contact
        subscription_form = ContactSubscriptionForm
    if not contact_list:
        raise Http404(
            'No ContactList or ContactQueue matches the given query.')

    data = {'properties': {}}
    if isinstance(contact_list, ContactQueue):
        data.update({'contact_queue': contact_list})
    else:
        data.update({'contact_list': contact_list})

    for field in request.POST:
        try:
            contact_model._meta.get_field(field)
            data[field] = request.POST.get(field)
        except FieldDoesNotExist:
            data['properties'][field] = request.POST.get(field)

    form = subscription_form(data=data)
    return render(
        request,
        'contact_queues/test_form.html',
        {'uuid': uuid, 'form': form})


@csrf_exempt
def subscription(request, uuid):
    """ Handle subscriptions from an external endpoint

    E.g. website of a organization

    That's why CSRF must be disabled on this view : there is no simple way the
    3rd-party website can produce a valid CSRF token in its form.
    """
    if request.method != 'POST':
        raise PermissionDenied('You cannot access this page directly.')

    contact_list = None
    try:
        contact_list = ContactQueue.objects.get(uuid=uuid)
        contact_model = CollectedContact
        subscription_form = CollectedContactSubscriptionForm
    except ContactQueue.DoesNotExist:
        contact_list = ContactList.objects.get(uuid=uuid)
        contact_model = Contact
        subscription_form = ContactSubscriptionForm
    if not contact_list:
        raise Http404(
            'No ContactList or ContactQueue matches the given query.')
    context = {}
    status_code = 201

    try:
        data = {'properties': {}}
        if isinstance(contact_list, ContactQueue):
            data.update({'contact_queue': contact_list})
        else:
            data.update({'contact_list': contact_list})

        for field in request.POST:
            try:
                contact_model._meta.get_field(field)
                data[field] = request.POST.get(field)
            except FieldDoesNotExist:
                data['properties'][field] = request.POST.get(field)

        form = subscription_form(data=data)
        if form.is_valid():
            contact = form.save(commit=False)
            if isinstance(contact_list, ContactQueue):
                cleaned_properties = PropertiesFieldHelper.to_internal_value(
                    data, reverse="contact_queue")
                contact.contact_queue = contact_list
            else:
                cleaned_properties = PropertiesFieldHelper.to_internal_value(
                    data, reverse="contact_list")
                contact.contact_list = contact_list
            contact.status = contact_model.PENDING
            contact.properties = cleaned_properties
            contact.subscription_ip = get_ip(request)
            contact.save()
            contact.apply_policies()
            context['contact'] = contact
        else:
            context['errors'] = form.errors
            status_code = 400
    except IntegrityError:
        context['errors'] = {
            '__all__': "Cette adresse est déjà en attente d'inscription."}
        status_code = 400
    except ValidationError as err:
        context['errors'] = err.detail
        status_code = 400

    return render(
        request, 'contacts/subscription.html', context, status=status_code)


def confirmation(request, uuid):
    expiration = timezone.now() -\
        settings.CONTACTS['EXPIRATIONS']['contact_queues:double-opt-in']
    filters = {
        'uuid': uuid,
        'update_date__gte': expiration,
        'status': AbstractContact.PENDING}
    contact = None
    try:
        contact = CollectedContact.objects.get(**filters)
    except CollectedContact.DoesNotExist:
        contact = Contact.objects.get(**filters)
    if not contact:
        raise Http404(
            'No Contact or CollectedContact matches the given query.')

    contact.status = AbstractContact.OK
    contact.save()
    return render(
        request, 'contacts/subscription.html', {'contact': contact})



import re
from email.utils import parseaddr

import dns
import dns.exception
import dns.resolver
import clamd
from django.conf import settings
from django.core.cache import cache
from django.core.validators import RegexValidator
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _

slug_regex = re.compile(r'^[a-z-]*$')
slug_regex_validator = RegexValidator(
    regex=slug_regex.pattern,
    message='source should be composed of lowercase letters and dashes')


def validate_existing_mail_domain(email):
    """ Just checks that the associated domain of an email has an MX record.
    """
    if settings.CAMPAIGNS['BYPASS_RECIPIENTS_MX_CHECK']:
        return
    domain = parseaddr(email)[1]

    if domain is None:
        # Trigger no error because email format is supposed to be checked
        # by another validator, needless to report it a second time
        return

    else:
        key = 'domcheck:{}'.format(domain)
        has_error = cache.get(key, None)

        if has_error is None:  # Not present in cache
            try:
                dns.resolver.query(domain, 'MX')
                has_error = False
            except dns.exception.DNSException:
                has_error = True
            finally:
                cache.set(key, has_error, 60)

        if has_error:
            raise ValidationError(_(
                'This domain "{}" doesn\'t seems to be '
                'configured to received emails').format(domain))


def validate_no_virus(afile):
    """ Checks a file field against a remote antivirus service
    """
    if not settings.CAMPAIGNS['SKIP_VIRUS_CHECK']:
        try:
            host, port = settings.CLAMD_HOST, settings.CLAMD_PORT
        except AttributeError:
            results = dns.resolver.query(settings.CLAMD_SERVICE_NAME, 'SRV')
            servers = [(rr.target.to_text(True), rr.port) for rr in results]
            host, port = servers[0]

        cd = clamd.ClamdNetworkSocket(host=host, port=port)
        check_result, check_msg = cd.instream(afile)['stream']
        # example virus result
        # {'stream': ('FOUND', 'Eicar-Test-Signature')}
        if check_result == 'FOUND':
            raise ValidationError(_(
                'File contains virus: "{}"').format(check_msg))
    return afile



from django.apps import AppConfig


class ContactsApp(AppConfig):
    name = 'munch.apps.contacts'
    verbose_name = 'Contacts'

    def ready(self):
        import munch.apps.contacts.urls  # noqa
        import munch.apps.contacts.api.v1.urls  # noqa



import unittest
import datetime

import pytz
from django.test import TestCase

from ..mail.utils import parse_email_date
from ..mail.utils import UniqueEmailAddressParser
from ..utils.permissions import MunchResourcePermission


class PermissionsTestCase(TestCase):
    def test_munch_resource_permission_names(self):
        self.assertEqual(
            MunchResourcePermission().mk_perms(verb='change'),
            (
                'campaigns.change_mine_mail',
                'campaigns.change_organizations_mail',
                'campaigns.change_mail'))



class TestMunchAddress(unittest.TestCase):
    def setUp(self):
        self.RPAddressParser = UniqueEmailAddressParser(
            domain='example.com', prefix='return-')

        self.UnsubscribeAddressParser = UniqueEmailAddressParser(
            domain='example.com', prefix='unsubscribe-')

    def test_valid_addr(self):
        s = 'return-fH7eGy3qQLuLGgr13RfJoA@example.com'
        self.assertTrue(
            self.RPAddressParser.is_valid(s))
        self.assertFalse(
            self.UnsubscribeAddressParser.is_valid(s))

    def test_invalid_addr(self):
        self.assertFalse(self.RPAddressParser.is_valid(
            'invalid+fH7eGy3qQLuLGgr13RfJoA@example.com'))

        self.assertFalse(self.RPAddressParser.is_valid(
            'return+fH7eGy3qQLuLGgr13RfJoA@invalid.example.com'))

    def test_creation(self):
        self.assertTrue(self.RPAddressParser.new())

    def test_creation_uuid_provided(self):
        ma = self.RPAddressParser.new('fH7eGy3qQLuLGgr13RfJoA')
        self.assertTrue(self.RPAddressParser.is_valid(ma))
        self.assertEqual(ma, 'return-fH7eGy3qQLuLGgr13RfJoA@example.com')

    def test_uuid_extraction(self):
        self.assertEqual(
            self.RPAddressParser.get_uuid(
                'return-fH7eGy3qQLuLGgr13RfJoA@example.com'),
            'fH7eGy3qQLuLGgr13RfJoA')

    def test_multiple_address_parser(self):
        RPAddressParser_1 = UniqueEmailAddressParser(
            domain='example_1', prefix='prefix_1')
        RPAddressParser_2 = UniqueEmailAddressParser(
            domain='example_2', prefix='prefix_2')

        self.assertNotEqual(RPAddressParser_1, RPAddressParser_2)
        self.assertNotEqual(RPAddressParser_1.domain, RPAddressParser_2.domain)
        self.assertNotEqual(RPAddressParser_1.prefix, RPAddressParser_2.prefix)


class TestUtils(unittest.TestCase):
    def test_parse_email_date(self):
        d = parse_email_date('Sat, 24 Nov 2035 11:45:15 -0500')
        self.assertEqual(
            pytz.utc.localize(datetime.datetime(
                year=2035, month=11, day=24, hour=16, minute=45, second=15)),
            d)



from rest_framework import viewsets
from rest_framework.response import Response

from munch.core.utils.views import NestedView
from munch.core.utils.views import MunchModelViewSetMixin
from munch.core.utils.views import OrganizationOwnedViewSetMixin

from ...models import SendingDomain
from .filters import SendingDomainFilter
from .serializers import SendingDomainSerializer


class SendingDomainViewSet(
        OrganizationOwnedViewSetMixin,
        MunchModelViewSetMixin,
        viewsets.ModelViewSet):
    model = SendingDomain
    serializer_class = SendingDomainSerializer
    filter_class = SendingDomainFilter

    def get_queryset(self):
        qs = super().get_queryset() | SendingDomain.objects.filter(
            alt_organizations__in=[self.request.user.organization])
        return qs.distinct()


class SendingDomainRevalidateView(NestedView):
    """ Allows to revalidate a domain

    Outputs the revalidation response, ex:
    `{"dkim_status":"ko"}`
    """
    parent_model = SendingDomain

    def post(self, request, pk):
        from ...tasks import validate_sending_domain_field

        domain = SendingDomain.objects.get(pk=pk)
        fields = ['dkim', 'app_domain']
        for field in fields:
            validate_sending_domain_field(pk, field)
        return Response({key: getattr(
            domain, '{}_status'.format(key)) for key in fields})



from django.test import TestCase
from django.conf import settings
from rest_framework.test import APIClient
from faker import Factory as FakerFactory

from munch.apps.users.tests.factories import UserFactory

from ...models import Mail
from ...models import Message
from ...tests.factories import MailFactory
from ...tests.factories import MessageFactory

faker = FakerFactory.create()


class MessageAPITestCase(TestCase):
    def setUp(self):
        self.api_prefix = 'v1'
        self.user = UserFactory(groups=['managers'])
        self.client = APIClient()
        self.client.login(identifier=self.user.identifier, password='password')

        self.html = '<html>{}</html>'.format(
            settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])

    def test_read_only_fields(self):
        response = self.client.post(
            '/{}/messages/'.format(self.api_prefix),
            {
                'name': 'test message',
                'subject': 'subject',
                'html': self.html,
                'sender_email': 'foo@bar.com',
                'sender_name': 'foo'}, format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Message.objects.count(), 1)

        url = response.data['url']

        response = self.client.put(
            url,
            {
                'name': 'test message',
                'subject': 'subject modified',
                'sender_email': 'foo@bar.com',
                'send_date': '1990-12-01',
                'sender_name': 'foo'}, format='json')
        self.assertEqual(response.json()['send_date'], None)
        self.assertEqual(response.json()['subject'], 'subject modified')

    def test_create_message(self):
        response = self.client.post(
            '/{}/messages/'.format(self.api_prefix),
            {
                'name': 'test message',
                'subject': 'subject',
                'html': self.html,
                'sender_email': 'foo@bar.com',
                'sender_name': 'foo'}, format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Message.objects.count(), 1)

        url = response.data['url']

        response = self.client.post(
            '/{}/recipients/'.format(self.api_prefix),
            {'message': url, 'to': 'foo@bar.com'}, format='json')
        response = self.client.patch(
            url, {'status': 'sending'}, format='json')
        self.assertEqual(response.json()['status'], 'sending')


class MailAPITestCase(TestCase):
    def setUp(self):
        self.api_prefix = 'v1'
        self.user = UserFactory(groups=['managers'])
        self.client = APIClient()
        self.client.login(identifier=self.user.identifier, password='password')

        self.html = '<html>{}</html>'.format(
            settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])

        self.message = MessageFactory(author=self.user)

    def test_empty_delete(self):
        MailFactory(message=self.message)
        MailFactory(message=self.message)
        response = self.client.delete(
            '/{}/recipients/'.format(self.api_prefix),
            [], format='json')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(Mail.objects.count(), 2)

        response = self.client.delete(
            '/{}/recipients/'.format(self.api_prefix),
            '', format='json')
        self.assertEqual(response.status_code, 400)

    def test_not_found_id(self):
        MailFactory(message=self.message)
        response = self.client.delete(
            '/{}/recipients/'.format(self.api_prefix),
            ['http://localhost:8000/v1/recipients/2'], format='json')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(Mail.objects.count(), 1)

    def test_delete_one(self):
        mail = MailFactory(message=self.message)
        response = self.client.get('/{}/recipients/{}/'.format(
            self.api_prefix, mail.id))
        response = self.client.delete(
            '/{}/recipients/'.format(self.api_prefix),
            [response.data['url']], format='json')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(Mail.objects.count(), 0)

    def test_delete_multiple(self):
        urls = []

        mail = MailFactory(message=self.message)
        response = self.client.get('/{}/recipients/{}/'.format(
            self.api_prefix, mail.id))
        urls.append(response.data['url'])
        mail = MailFactory(message=self.message)
        response = self.client.get('/{}/recipients/{}/'.format(
            self.api_prefix, mail.id))
        urls.append(response.data['url'])

        self.assertEqual(Mail.objects.count(), 2)
        response = self.client.delete(
            '/{}/recipients/'.format(self.api_prefix), urls, format='json')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(Mail.objects.count(), 0)

    def test_delete_multiple_with_one_not_found(self):
        urls = ['http://localhost:8000/v1/recipients/12321321/']

        mail = MailFactory(message=self.message)
        response = self.client.get('/{}/recipients/{}/'.format(
            self.api_prefix, mail.id))
        urls.append(response.data['url'])
        mail = MailFactory(message=self.message)
        response = self.client.get('/{}/recipients/{}/'.format(
            self.api_prefix, mail.id))
        urls.append(response.data['url'])

        self.assertEqual(Mail.objects.count(), 2)
        response = self.client.delete(
            '/{}/recipients/'.format(self.api_prefix), urls, format='json')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(Mail.objects.count(), 0)

    def test_delete_not_owned(self):
        user = UserFactory(groups=['managers'])
        message = MessageFactory(author=user)
        mail = MailFactory(message=message)

        self.client.login(identifier=user.identifier, password='password')
        response = self.client.get('/{}/recipients/{}/'.format(
            self.api_prefix, mail.id))

        self.client.login(identifier=self.user.identifier, password='password')
        response = self.client.delete(
            '/{}/recipients/'.format(self.api_prefix),
            [response.data['url']], format='json')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(Mail.objects.count(), 1)


class MailBulkAPITestCase(TestCase):
    def setUp(self):
        self.api_prefix = 'v1'
        self.user = UserFactory(groups=['managers'])
        self.client = APIClient()
        self.client.login(identifier=self.user.identifier, password='password')

        self.html = '<html>{}</html>'.format(
            settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])

        self.message = MessageFactory(author=self.user)
        self.message_url = self.client.get('/{}/messages/{}/'.format(
            self.api_prefix, self.message.id)).data['url']

    def test_bad_format(self):
        email = faker.email()

        data = [
            [{'message': self.message_url, 'to': email}],
            {'message': self.message_url, 'to': ''}
        ]
        response = self.client.post(
            '/{}/recipients/'.format(self.api_prefix), data, format='json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.json()['duplicates'], [])
        self.assertEqual(response.json()['validation_errors'], {})
        self.assertEqual(response.json()['no_address'], 2)
        self.assertEqual(len(response.json()['results']), 0)

        data = [
            'Jambon',
            {'message': self.message_url, 'to': email}
        ]
        response = self.client.post(
            '/{}/recipients/'.format(self.api_prefix), data, format='json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.json()['duplicates'], [])
        self.assertEqual(response.json()['validation_errors'], {})
        self.assertEqual(response.json()['no_address'], 1)
        self.assertEqual(len(response.json()['results']), 0)

    def test_missing_address(self):
        email = faker.email()

        data = [
            {'message': self.message_url, 'to': email},
            {'message': self.message_url, 'to': ''}
        ]
        response = self.client.post(
            '/{}/recipients/'.format(self.api_prefix), data, format='json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.json()['duplicates'], [])
        self.assertEqual(response.json()['validation_errors'], {})
        self.assertEqual(response.json()['no_address'], 1)
        self.assertEqual(len(response.json()['results']), 1)

        recipient = response.json()['results'][0]
        self.assertEqual(recipient['to'], email)
        self.assertEqual(Mail.objects.count(), 1)

    def test_duplicates(self):
        email = faker.email()

        data = [
            {'message': self.message_url, 'to': faker.email()},
            {'message': self.message_url, 'to': email},
            {'message': self.message_url, 'to': email}
        ]
        response = self.client.post(
            '/{}/recipients/'.format(self.api_prefix), data, format='json')
        self.assertEqual(response.json()['duplicates'], [email])
        self.assertEqual(len(response.json()['results']), 2)
        self.assertEqual(Mail.objects.count(), 2)

    def test_unicity(self):
        email = faker.email()
        MailFactory(recipient=email, message=self.message)

        data = [
            {'message': self.message_url, 'to': faker.email()},
            {'message': self.message_url, 'to': email}
        ]
        response = self.client.post(
            '/{}/recipients/'.format(self.api_prefix), data, format='json')
        self.assertEqual(response.json()['duplicates'], [])
        self.assertEqual(len(response.json()['results']), 1)
        self.assertIn(email, response.json()['validation_errors'])
        self.assertEqual(
            response.json()['validation_errors'][email],
            ['This address is already attached to this message.'])
        self.assertEqual(Mail.objects.count(), 2)

    def test_all_recipients_must_have_same_message(self):
        message = MessageFactory(author=self.user)
        message_url = self.client.get('/{}/messages/{}/'.format(
            self.api_prefix, message.id)).data['url']

        data = [
            {'message': self.message_url, 'to': faker.email()},
            {'message': message_url, 'to': faker.email()}
        ]
        response = self.client.post(
            '/{}/recipients/'.format(self.api_prefix), data, format='json')
        self.assertEqual(response.json()['results'], [])
        self.assertEqual(
            response.json()['non_field_errors'],
             ['Every object must have same message'])
        self.assertEqual(Mail.objects.count(), 0)

    def test_another_error(self):
        email = faker.email()
        data = [{'message': 'http://dadada', 'to': email}]

        response = self.client.post(
            '/{}/recipients/'.format(self.api_prefix), data, format='json')
        self.assertEqual(response.json()['duplicates'], [])
        self.assertIn(email, response.json()['validation_errors'])
        self.assertIn('message', response.json()['validation_errors'][email])



import sys

from celery.signals import celeryd_after_setup

from munch.core.utils import get_worker_type
from munch.core.celery import catch_exception
from munch.core.celery import munch_tasks_router


def register_tasks():
    tasks_map = {
        'status': [
            'munch.apps.transactional.status.SendWebhook',
            'munch.apps.transactional.status.CreateDSN',
            'munch.apps.transactional.status.ForwardDSN',
            'munch.apps.transactional.status.HandleDSNStatus',
            'munch.apps.transactional.status.HandleSMTPStatus',
            'munch.apps.transactional.status.record_status'
        ],
    }
    munch_tasks_router.import_tasks_map(tasks_map, 'transactional')


@celeryd_after_setup.connect
@catch_exception
def configure_worker(instance, **kwargs):
    if get_worker_type() in ['status', 'all']:
        from .status import create_dsn  # noqa
        from .status import forward_dsn  # noqa
        from .status import send_webhook  # noqa
        from .status import record_status  # noqa
        from .status import handle_dsn_status  # noqa
        from .status import handle_smtp_status  # noqa
        sys.stdout.write('[transactional] Registering worker as STATUS...')
        munch_tasks_router.register_as_worker('status')



import string
from random import choice

from django.db import models
from django.db import transaction
from django.conf import settings
from django.contrib.auth.models import BaseUserManager
from django.contrib.auth.models import AbstractBaseUser
from django.contrib.auth.models import PermissionsMixin
from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django.utils import timezone
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from django.utils.translation import ugettext_lazy as _
from django.template.loader import render_to_string

from munch.core.mail.utils import mk_base64_uuid
from munch.core.mail.utils.emails import ServiceMessage
from munch.core.utils import get_login_url
from munch.core.utils.models import AbstractOwnedModel
from munch.core.utils.managers import OwnedModelQuerySet
from munch.core.utils.models import OwnedModelMixin
from munch.core.models import ValidationSignalsModel

from .tokens import token_generator


class Organization(ValidationSignalsModel, AbstractOwnedModel):
    name = models.CharField(
        _('name'), max_length=200, help_text=_("Organization name"))
    contact_email = models.EmailField(
        _('contact email address'),
        help_text=_(
            'Main contact email for organization. '
            'Will also receive status/optouts notifications'))
    can_external_optout = models.BooleanField(
        _('Can create messages with external optout'), default=False)
    can_attach_files = models.BooleanField(
        _('Can add attachments'), default=False)
    parent = models.ForeignKey(
        'self', null=True, blank=True, related_name='children')
    creation_date = models.DateTimeField(default=timezone.now)
    update_date = models.DateTimeField(blank=True)

    class Meta(OwnedModelMixin.Meta):
        verbose_name = _('organization')
        verbose_name_plural = _('organizations')

    owner_path = 'self'
    author_path = AbstractOwnedModel.IRRELEVANT

    def __str__(self):
        return '{}'.format(self.name)

    def get_organization(self):
        return self

    def validate_children(self):
        if self.parent:
            if self.parent == self:
                raise ValidationError(
                    'Cannot set myself as parent.')
            if self.children.exists():
                raise ValidationError(
                    'Cannot have a parent and children at same time.')
            if self.parent.parent:
                raise ValidationError(
                    'Cannot be attached to a parent '
                    'organization that already has a parent.')

    def save(self, *args, **kwargs):
        created = not self.pk
        self.update_date = timezone.now()
        super().save(*args, **kwargs)
        if created:
            os = OrganizationSettings(organization=self)
            os.save()

    def get_all_users(self):
        return MunchUser.objects.filter(
            organization=self) | MunchUser.objects.filter(
                organization__in=self.children.all())


class OrganizationSettings(AbstractOwnedModel):
    nickname_valid_msg = _(
        'Nickname can only contain lowercase letters, '
        'numbers and dashes. It must be between 4 and 30 characters long')

    organization = models.OneToOneField(
        'Organization', related_name='settings',
        verbose_name=_('organization'))
    nickname = models.CharField(
        _('nickname'), max_length=30, blank=True, default='',
        help_text=_("Organization nickname for tracking links"),
        validators=[RegexValidator(
            '^[a-z0-9-]{4,30}$', message=nickname_valid_msg)])

    notify_message_status = models.BooleanField(
        _('Receive message status notification emails'), default=True)
    notify_optouts = models.BooleanField(
        _('Receive optouts notification emails'), default=False)
    external_optout_message = models.TextField(
        _("External optout contact"), blank=True, default='',
        help_text=_(
            "Text which will be display to recipients who want to contact you "
            "in order to unsubscribe in case of external optout"))

    owner_path = 'organization'
    author_path = AbstractOwnedModel.IRRELEVANT

    class Meta(AbstractOwnedModel.Meta):
        verbose_name = _("organization settings")
        verbose_name_plural = _("organization settings")

    def clean(self, *args, **kwargs):
        # a bit of duplicate with OrganizationSettingsSerializer.validate
        if (not self.organization.can_external_optout and
                self.external_optout_message.strip()):
            raise ValidationError(_(
                'External optout is not allowed, no need '
                'to provide a message'))


class MunchUserManager(BaseUserManager):
    def _create_user(self, identifier, password, is_superuser, **extra_fields):
        """
        Creates and saves a User with the given username, email and password.
        """
        if not identifier:
            raise ValueError(_('The given identifier must be set'))
        identifier = self.normalize_email(identifier)
        user = self.model(
            identifier=identifier, is_active=True,
            is_superuser=is_superuser, **extra_fields)
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_superuser(self, identifier, password, *args, **kwargs):
        return self._create_user(
            identifier, password, is_superuser=True, *args, **kwargs)


class MunchUser(
        ValidationSignalsModel, AbstractBaseUser,
        PermissionsMixin, OwnedModelMixin):
    """
    Simple user model for munch users
    Includes both admins and regular users.
    """
    SECRET_LENGTH = 30

    identifier = models.EmailField(
        _('identifier'), unique=True, blank=False,
        null=False, help_text=_('identifier'))
    first_name = models.CharField(_('first name'), max_length=50, blank=False)
    last_name = models.CharField(_('last name'), max_length=50, blank=False)
    organization = models.ForeignKey(
        'Organization', verbose_name=_('organization'),
        null=True, related_name='users')
    is_active = models.BooleanField(_('active'), default=False)
    is_admin = models.BooleanField(_('admin'), default=False)

    secret = models.CharField(_('secret'), max_length=SECRET_LENGTH)

    creation_date = models.DateTimeField(_('creation date'), auto_now_add=True)

    invited_by = models.ForeignKey(
        'MunchUser', verbose_name=_('invited by'), null=True, blank=True,
        on_delete=models.SET_NULL)

    USERNAME_FIELD = 'identifier'

    objects = MunchUserManager.from_queryset(OwnedModelQuerySet)()

    owner_path = 'organization'
    author_path = 'self'

    class Meta(OwnedModelMixin.Meta):
        verbose_name = _('user')
        verbose_name_plural = _('users')

    def save(self, *args, **kwargs):
        created = False
        if not self.pk:
            created = True

        # If no API key exists, just produce one.
        # Typically: at instance creation
        with transaction.atomic():
            if not self.secret:
                self.regen_secret()
            super().save(*args, **kwargs)

        if created and not self.password:
            transaction.on_commit(self.send_invitation_email)

    def regen_secret(self):
        self.secret = self._mk_random_secret()

    def _send_service_message(
            self, subject, template, template_html, extra_context={}):
        protocol, domain = settings.HELPDESK_URL.strip('/').split('://')
        context = {
            'identifier': self.identifier,
            'user_name': self.get_short_name(),
            'product_name': settings.PRODUCT_NAME,
            'protocol': protocol,
            'domain': domain,
            'user': self,
        }
        context.update(extra_context)
        message = ServiceMessage(
            subject=subject, to=self.identifier, template=template,
            render_context=context)
        message.add_html_part_if_exists(template_html)
        message.send()

    def send_invitation_email(self):
        self._send_service_message(
            subject=render_to_string(
                'users/emails/invitation_email_subject.txt', {
                    'product_name': settings.PRODUCT_NAME,
                    'user_name': self.get_short_name()}).strip(),
            template='users/emails/invitation_email.txt',
            template_html='users/emails/invitation_email.html',
            extra_context={
                'token': token_generator.make_token(self),
                'uid': urlsafe_base64_encode(force_bytes(self.pk))}
        )

    def send_invitation_complete_email(self):
        self._send_service_message(
            subject=render_to_string(
                'users/emails/invitation_complete_email_subject.txt', {
                    'product_name': settings.PRODUCT_NAME,
                    'user_name': self.get_short_name()}).strip(),
            template='users/emails/invitation_complete_email.txt',
            template_html='users/emails/invitation_complete_email.html',
            extra_context={'login_url': get_login_url()}
        )

    def send_password_reset_email(self):
        self._send_service_message(
            subject=render_to_string(
                'users/emails/password_reset_email_subject.txt', {
                    'product_name': settings.PRODUCT_NAME,
                    'user_name': self.get_short_name()}).strip(),
            template='users/emails/password_reset_email.txt',
            template_html='users/emails/password_reset_email.html',
            extra_context={
                'token': token_generator.make_token(self),
                'uid': urlsafe_base64_encode(force_bytes(self.pk))}
        )

    def send_password_reset_complete_email(self):
        self._send_service_message(
            subject=render_to_string(
                'users/emails/password_reset_complete_email_subject.txt', {
                    'product_name': settings.PRODUCT_NAME,
                    'user_name': self.get_short_name()}).strip(),
            template='users/emails/password_reset_complete_email.txt',
            template_html='users/emails/password_reset_complete_email.html',
            extra_context={'login_url': get_login_url()}
        )

    @property
    def full_name(self):
        return self.get_full_name()
    full_name.fget.short_description = _('name')

    def get_full_name(self):
        if self.first_name or self.last_name:
            return '{} {}'.format(self.first_name, self.last_name)
        else:
            return self.identifier

    def get_short_name(self):
        return self.first_name or self.identifier

    def __str__(self):
        return self.identifier

    @property
    def is_staff(self):
        return self.is_admin

    def has_perm(self, perm, obj=None):
        if self.is_admin:
            return True
        return super().has_perm(perm, obj)

    def has_module_perms(self, app_label):
        return True

    @classmethod
    def _mk_random_secret(cls, length=None):
        length = length or cls.SECRET_LENGTH
        chars = string.ascii_letters + string.digits
        return ''.join(choice(chars) for _ in range(length))


class APIApplication(AbstractOwnedModel):
    identifier = models.CharField(
        max_length=50, blank=False, null=False, verbose_name=_('identifier'))
    secret = models.CharField(
        _('secret'), max_length=25, default=mk_base64_uuid)
    author = models.ForeignKey(MunchUser, verbose_name=_('author'))

    owner_path = 'author__organization'
    author_path = 'author'

    class Meta:
        verbose_name = _('api application')
        verbose_name_plural = _('api applications')
        unique_together = [('identifier', 'author')]

    def __str__(self):
        return '{} ({})'.format(self.identifier, self.author)

    def regen_secret(self):
        self.secret = mk_base64_uuid()

    def save(self, *args, **kwargs):
        with transaction.atomic():
            if not self.secret:
                self.regen_secret()
            super().save(*args, **kwargs)


class SmtpApplication(AbstractOwnedModel):
    identifier = models.CharField(
        max_length=50, blank=False, null=False, verbose_name=_('identifier'))
    username = models.CharField(
        max_length=25, db_index=True, unique=True,
        default=mk_base64_uuid, verbose_name=_('username'))
    secret = models.CharField(
        _('secret'), max_length=25, default=mk_base64_uuid)
    author = models.ForeignKey(MunchUser, verbose_name=_('author'))

    owner_path = 'author__organization'
    author_path = 'author'

    class Meta:
        verbose_name = _('smtp application')
        verbose_name_plural = _('smtp applications')
        unique_together = [('identifier', 'author')]

    def __str__(self):
        return '{} ({})'.format(self.identifier, self.author)

    def regen_credentials(self):
        self.username = mk_base64_uuid()
        self.secret = mk_base64_uuid()

    def save(self, *args, **kwargs):
        with transaction.atomic():
            if not self.secret or not self.username:
                self.regen_credentials()
            super().save(*args, **kwargs)



import functools

import celery
import celery.exceptions
from celery import task
from celery.exceptions import Reject
from celery.exceptions import IncompleteStream
from celery.utils.log import get_task_logger

log = get_task_logger(__name__)


class AutoRetryTask(celery.task.Task):
    """ A task that will retry automatically on exception

    exceptions triggering retry are all exceptions, excepted celery ones.

    """
    # Builds a list of celery exceptions, excluding warnings
    # workaround https://github.com/celery/celery/pull/2644
    CELERY_EXCEPTIONS = tuple(
        v for k, v in celery.exceptions.__dict__.items()
        if ((k in celery.exceptions.__all__) and (not k.endswith('Warning')))
    ) + (Reject, IncompleteStream)

    def run(self, *args, **kwargs):
        try:
            self.safe_run(*args, **kwargs)
        except Exception as exc:

            if isinstance(exc, self.CELERY_EXCEPTIONS):
                raise
            else:
                log.exception(exc)
                self.retry(exc=exc)

    def safe_run(self, *args, **kwargs):
        raise NotImplementedError


def task_autoretry(*args_task, **kwargs_task):
    # https://github.com/celery/celery/pull/2112
    def real_decorator(func):
        @task(*args_task, **kwargs_task)
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except kwargs_task.get('autoretry_on', Exception) as exc:
                for exclude in kwargs_task.get('autoretry_exclude', []):
                    if isinstance(exc, exclude):
                        log.info(
                            'Wont retry this task because exception '
                            '"{}" is exclude'.format(str(exc)))
                        return
                if kwargs_task.get('retry_message', False):
                    log.error(kwargs_task.get('retry_message'), exc_info=True)
                wrapper.retry(exc=exc)
        return wrapper
    return real_decorator



import os
import logging

from django.conf import settings
from slimta.policy import QueuePolicy

logger = logging.getLogger(__name__)


class Apply(QueuePolicy):
    """
    Execute abritary policies on file system based on EXEC_QUEUE_POLICIES
    """
    def apply(self, envelope):
        for path in settings.TRANSACTIONAL.get('EXEC_QUEUE_POLICIES'):
            if os.path.exists(path):
                with open(path) as module:
                    ephemeral_context = {}
                    allowed_context = {'__builtins__': {
                        'settings': settings,
                        'print': print,
                        'logger': logger,
                        'math': __import__('math')}}
                    try:
                        exec(module.read(), allowed_context, ephemeral_context)
                        ephemeral_context.get('apply')(envelope)
                    except Exception as err:
                        logger.warning(
                            '[{}] Failed to execute "{}" ephemeral '
                            'policy: {}'.format(
                                envelope.headers.get(
                                    settings.TRANSACTIONAL.get(
                                        'X_MESSAGE_ID_HEADER',
                                        'NO-MESSAGE-ID')),
                                path, err))
            else:
                logger.warning(
                    "[{}] Following ephemeral policy doesn't "
                    "exists: {}".format(
                        envelope.headers.get(settings.TRANSACTIONAL.get(
                            'X_MESSAGE_ID_HEADER', 'NO-MESSAGE-ID')),
                        path))



import logging

from django.conf import settings
from slimta.policy import RelayPolicy

from munch.core.mail.utils import UniqueEmailAddressParser

log = logging.getLogger(__name__)

return_path_parser = UniqueEmailAddressParser(
    domain=lambda: settings.RETURNPATH_DOMAIN, prefix='return-')


class RewriteReturnPath(RelayPolicy):
    """ Rewrite existing ReturnPath based on some conditions """
    def apply(self, envelope):
        # If there is X-HTTP-Return-Path
        if settings.TRANSACTIONAL['X_HTTP_DSN_RETURN_PATH_HEADER'] in \
                envelope.headers:
            # Rewrite it
            envelope.sender = return_path_parser.new(
                envelope.headers[settings.TRANSACTIONAL[
                    'X_MESSAGE_ID_HEADER']])
        # If there is no X-HTTP-Return-Path
        else:
            log.info(
                '[{}] Keeping original Return-Path ({}) because there is no '
                '"{}" (then DSN will be sent to original sender).'.format(
                    envelope.headers[settings.TRANSACTIONAL[
                        'X_MESSAGE_ID_HEADER']],
                    envelope.sender,
                    settings.TRANSACTIONAL['X_HTTP_DSN_RETURN_PATH_HEADER']))



import logging

from django.conf import settings
from django.dispatch import receiver
from django.db.models.signals import pre_save
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext as _

from munch.apps.campaigns.models import Message

from .utils import SPFValidator
from .models import SendingDomain


log = logging.getLogger(__name__)


@receiver(pre_save, sender=Message)
def validate_from_domain(sender, instance, raw, *args, **kwargs):
    if raw:
        return

    if settings.BYPASS_DNS_CHECKS:
        log.debug(
            'Bypassing dns checks because it '
            'settings "BYPASS_DNS_CHECKS" is enabled.')
        return

    if instance.pk:
        previous = Message.objects.only('sender_email').get(pk=instance.id)
        if previous.sender_email == instance.sender_email:
            log.info(
                'Bypassing dns checks because '
                '"Message.sender_email" has not changed.')
            return

    message = instance
    try:
        all_domains = SendingDomain.objects
        org_id = message.get_organization()
        qs = all_domains.filter(organization=org_id) | all_domains.filter(
            alt_organizations__in=[org_id])
        organization_domains = qs.distinct()
        domain = organization_domains.get_from_email_addr(message.sender_email)
    except SendingDomain.DoesNotExist:
        raise ValidationError(_(
            'You must add and configure "{}" domain'.format(
                message.sender_email)))
    if not domain.is_valid():
        raise ValidationError(_(
            'Sending domain invalid, please verify '
            'DNS configuration ({})'.format(domain.name)))


@receiver(pre_save, sender=Message)
def validate_envelope_domain(sender, instance, raw, *args, **kwargs):
    if raw or settings.BYPASS_DNS_CHECKS:
        return
    if instance.status == Message.SENDING:
        previous = Message.objects.only('status').get(pk=instance.pk)
        if previous.status != Message.SENDING:
            validator = SPFValidator(
                spf_include=settings.DOMAINS['SPF_INCLUDE'])
            validator.validate(settings.RETURNPATH_DOMAIN)



import re
import time
import email
import logging
import datetime
from email.utils import parseaddr
from email.utils import formatdate

import celery
import requests
from django.conf import settings
from slimta.envelope import Envelope

from munch.core.mail.backend import Backend
from munch.core.mail.utils import extract_domain
from munch.core.mail.utils import parse_email_date
from munch.core.utils.tasks import task_autoretry
from munch.core.mail.utils.dsn import DSN
from munch.core.mail.utils.dsn import DSNParser
from munch.core.mail.utils.dsn import InvalidDSNError
from munch.core.mail.exceptions import SoftFailure
from munch.core.mail.exceptions import HardFailure
from munch.core.mail.exceptions import RejectMissingField
from munch.core.mail.exceptions import RejectInvalidDSN

from .models import Mail
from .models import MailStatus
from .utils import get_envelope_from_identifier
from .policies.relay.headers import return_path_parser

log = logging.getLogger(__name__)


class SendWebhook(celery.task.Task):
    max_retries = None

    # SMTP status map
    # It can be used with first digit of both ESMTP and SMTP status codes
    @staticmethod
    def get_http_returnpath(identifier, headers):
        """ Fetch the http stored returnpath

        See http_returnpath policy for the inverse action (storing)

        :param: the CM-Message-Id mail address of just sent envelope
        :return: the URL to notify status updates
        """
        mail = Mail.objects.get(identifier=identifier)
        http_return_path = mail.headers.get(
            settings.X_HTTP_DSN_RETURN_PATH_HEADER, None)
        if http_return_path:
            return http_return_path.strip()

    def http_post(self, url, data):
        tmp_err_msg = None

        try:
            response = requests.post(url, json=data)
        except requests.exceptions.ConnectionError as e:
            tmp_err_msg = (
                'Failed to log status "{}" to {} : {}'.format(
                    data['status'], url, str(e)))
        else:
            if response.status_code in (200, 201):
                log.info('Logged "{}" status to {}'.format(
                    data['status'], url))
            else:
                tmp_err_msg = (
                    'Failed to log status "{}" to {} (HTTP {})'.format(
                        data['status'], url, response.status_code))
        finally:
            if tmp_err_msg is not None:
                try:
                    self.retry(
                        countdown=settings.TRANSACTIONAL[
                            'STATUS_WEBHOOK_RETRY_INTERVAL'],
                        max_retries=settings.TRANSACTIONAL[
                            'STATUS_WEBHOOK_RETRIES'])
                except celery.exceptions.MaxRetriesExceededError:
                    log.warn(
                        'Too many errors while sending webhook, giving up.',
                        exc_info=True, extra={'error_message': tmp_err_msg})
                except celery.exceptions.Retry:
                    # In case we just queued a retry task
                    log.info(
                        'Error while sending webhook (will retry later)',
                        exc_info=True, extra={'error_message': tmp_err_msg})
                    raise

    def log_no_http_returnpath(self, identifier):
        # may become log.info() once we handle properly the sending of
        # bounces via SMTP
        log.info(
            '[{}] No {}, unable to HTTP-push the status'.format(
                identifier, settings.X_HTTP_DSN_RETURN_PATH_HEADER))

    def __call__(self, identifier, headers, data):
        webhook_url = self.get_http_returnpath(identifier, headers)
        if not webhook_url:
            self.log_no_http_returnpath(identifier)
        else:
            self.http_post(webhook_url, data)


send_webhook = SendWebhook()


class SendDSN(celery.task.Task):
    @staticmethod
    def get_smtp_returnpath(identifier, headers, mail=None):
        """ Fetch the smtp stored returnpath

        See smtp_returnpath policy for the inverse action (storing)

        :param: the Return-Path mail address of just sent envelope
        :return: the URL to notify status updates
        """
        if not mail:
            mail = Mail.objects.get(identifier=identifier)
        if mail.headers.get(
                settings.X_HTTP_DSN_RETURN_PATH_HEADER, None):
            return mail.headers.get(
                settings.X_SMTP_DSN_RETURN_PATH_HEADER, None)
        else:
            return mail.headers.get(
                settings.X_SMTP_DSN_RETURN_PATH_HEADER, mail.sender)

    def send_dsn(self, envelope, recipient):
        backend = Backend()

        # We set an empty MAILFROM since DSN don't have a Return-Path
        envelope.sender = ''
        envelope.recipients = [recipient]
        backend.send_simple_envelope(envelope)


class CreateDSN(SendDSN):
    max_retries = None

    def __call__(self, identifier, headers, client, reply, recipient,
                 status_date, reporting_mta):

        body = "[message omitted] [{}]".format(identifier).encode('utf-8')
        remote_mta = False
        if reply.address:
            remote_mta = reply.address[0]
        dsn = DSN(dsn_from=settings.TRANSACTIONAL.get('SMTP_DSN_ADDRESS'),
                  dsn_to=recipient)

        # convert status_date (datetime) to a string usable in DSN
        timeval = time.mktime(status_date.timetuple())
        dsn_date = formatdate(timeval=timeval, localtime=True)
        dsn_message = dsn.generate(
            original_headers=headers, original_body=body,
            reply_code=reply.code, reply_message=reply.message,
            reporting_mta=reporting_mta, remote_mta=remote_mta, date=dsn_date,
            report_headers={
                settings.TRANSACTIONAL['X_MESSAGE_ID_HEADER']: identifier})

        envelope = Envelope()
        envelope.parse(dsn_message)

        self.send_dsn(envelope, recipient)


create_dsn = CreateDSN()


class ForwardDSN(SendDSN):
    max_retries = None

    def __call__(self, original_dsn, recipient):

        envelope = Envelope()
        envelope.parse(original_dsn.encode('utf-8'))

        self.send_dsn(envelope, recipient)


forward_dsn = ForwardDSN()


class HandleDSNStatus(celery.task.Task):
    max_retries = None

    RETURNPATH_ADDR = 'To'
    DATE_FIELD = 'Arrival-Date'

    def __call__(self, original_dsn, report_headers):
        original_dsn = email.message_from_string(original_dsn)
        try:
            return_path = parseaddr(original_dsn[self.RETURNPATH_ADDR])[1]
            message_id = return_path_parser.get_uuid(return_path)
        except (KeyError, ValueError):
            raise RejectMissingField(
                self.RETURNPATH_ADDR, 'unknown', headers=original_dsn)

        try:
            dsn = DSNParser(report_headers)
        except InvalidDSNError as e:
            raise RejectInvalidDSN(e, headers=report_headers)

        try:
            # 2 fallbacks for the date
            if dsn.date:
                date = dsn.date
            else:
                if 'Date' in original_dsn:
                    date = parse_email_date(original_dsn['Date'])
                else:
                    date = datetime.datetime.now()

            status = dsn.get_next_mailstatus()

            # Record status
            if dsn.original_recipient:
                recipient = dsn.original_recipient
            else:
                recipient = dsn.final_recipient

            original_domain = extract_domain(recipient)

            # We use a generic source_ip for this status
            # since it is not linked to one of our workers
            raw_msg = ''
            if dsn.smtp_status:
                raw_msg = dsn.smtp_status + ' '
            if dsn.msg:
                raw_msg += dsn.msg.replace('\n', ' ')
            mailstatus = MailStatus(
                status_code=dsn.esmtp_status,
                raw_msg=raw_msg.strip(),
                status=status, destination_domain=original_domain,
                source_ip='0.0.0.0')
            record_status(mailstatus, message_id, '0.0.0.0')

            # Send Webhook
            data = {
                'status': status,
                'message': dsn.msg.replace('\n', ' '),
                'smtp_status': dsn.smtp_status or 'unknown',
                'esmtp_status': dsn.esmtp_status,
                'date': date.isoformat(),
                'recipient': recipient
            }
        except Exception as exc:
            log.error(
                'Error while handling DSN Status. Retrying.', exc_info=True)
            self.retry(exc=exc)

        send_webhook.apply_async(
            args=[message_id, original_dsn.as_string(), data])

        # Forward DSN by mail
        # Only if status is Bounced or Dropped
        if status in [MailStatus.BOUNCED, MailStatus.DROPPED]:
            smtp_return_path = SendDSN.get_smtp_returnpath(
                message_id, original_dsn)
            if smtp_return_path:
                # First, replace the the DSN To: (our return address) by the
                # original sender so that it appears properly
                original_dsn.replace_header('To', smtp_return_path)
                forward_dsn.apply_async(
                    args=[original_dsn.as_string(), smtp_return_path])

        else:
            log.info(
                "[{}] Won't forward this DSN because it's "
                "not a failure one.".format(message_id))


handle_dsn_status = HandleDSNStatus()


class HandleSMTPStatus(celery.task.Task):
    max_retries = None

    REGEXP_ESMTP_STATUS = re.compile(
        r'^(?P<esmtp_status>\d+\.\d+\.\d+)\s+(?P<message>.*)')

    @classmethod
    def _extract_esmtp_status(cls, reply):
        """
        :type reply: slimta.smtp.reply.Reply
        :return: a ESMTP status code (ex: "2.0.0") or "unknown"
        :rtype str:
        """
        m = cls.REGEXP_ESMTP_STATUS.match(reply.message)
        if m:
            return m.group('esmtp_status')
        else:
            return 'unknown'

    def __call__(self, status, status_date, identifier, reply, relay_ehlo):
        try:
            envelope = get_envelope_from_identifier(
                identifier, only_headers=True)

            webhook_data = {
                'status': status,
                'message': reply.message,
                'smtp_status': reply.code,
                'esmtp_status': self._extract_esmtp_status(reply),
                'date': status_date.isoformat(),
                'recipient': envelope.recipients[0]}
        except Exception as exc:
            log.error(
                'Error while handling SMTP Status. Retrying.', exc_info=True)
            self.retry(exc=exc)

        # Send Webhook
        send_webhook.apply_async(
            [identifier, dict(envelope.headers), webhook_data])

        # Send DSN by mail
        # Only if status is Bounced or Dropped
        if status in [MailStatus.BOUNCED, MailStatus.DROPPED]:
            smtp_return_path = SendDSN.get_smtp_returnpath(
                identifier, envelope.headers)
            if smtp_return_path:
                create_dsn.apply_async([
                    identifier, dict(envelope.headers), envelope.client,
                    reply, smtp_return_path, status_date, relay_ehlo])

        else:
            log.info(
                "[{}] Won't forward this DSN because it's "
                "not a failure one.".format(identifier))


handle_smtp_status = HandleSMTPStatus()


@task_autoretry(
    max_retries=None, autoretry_on=(Exception, ),
    autoretry_exclude=(SoftFailure, HardFailure, ),
    retry_message='Error while trying to record_status. Retrying.')
def record_status(mailstatus, identifier, relay_ehlo, reply=None):
    """ Save MailStatus and send SMTP Status task """
    log.info('[{}] Recording "{}" status...'.format(
        identifier, mailstatus.status))
    try:
        mail = Mail.objects.get(identifier=identifier)
        mailstatus.mail = mail
        mailstatus.save()
        if reply:
            handle_smtp_status.apply_async([
                mailstatus.status, mailstatus.creation_date, identifier,
                reply, relay_ehlo])
    except (SoftFailure, Mail.DoesNotExist) as exc:
        raise SoftFailure(exc)



from django.shortcuts import get_object_or_404
from django.http import HttpResponse

from .munchers import post_web_html_individual
from munch.apps.campaigns.models import Message
from munch.apps.tracking.views import web_tracking_open


def hosted_message(request, identifier):
    msg = get_object_or_404(Message, identifier=identifier)

    # Handles tracking  (optionally)
    mail = web_tracking_open(request, msg)

    if mail:
        html = post_web_html_individual.process(
            msg.html,
            app_url=msg.get_app_url(),
            track_clicks=msg.track_clicks,
            unsubscribe_url=mail.unsubscribe_url,
            mail_identifier=mail.identifier,
            links_map=msg.msg_links)
    else:
        html = msg.html
    return HttpResponse(html)



from django.test import TestCase

from munch.apps.campaigns.models import OptOut
from munch.apps.campaigns.tests.factories import MailFactory
from munch.apps.campaigns.tests.factories import MessageFactory
from munch.apps.campaigns.tests.factories import PreviewMailFactory
from munch.apps.users.tests.factories import UserFactory

from .factories import OptOutFactory


class TestOptOut(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.message = MessageFactory(author=self.user)

    def test_new_optout(self):
        """ OptOut when no other opptout exist for this email"""
        mail = MailFactory(message=self.message)

        self.assertFalse(OptOut.objects.filter(
            identifier=mail.identifier).exists())
        response = self.client.post(
            '/h/subscriptions/{}/optout/'.format(mail.identifier))
        self.assertEqual(response.status_code, 302)

        o = OptOut.objects.get(identifier=mail.identifier)
        self.assertEqual(o.origin, OptOut.BY_WEB)

    def test_optout_again(self):
        """ OptOut when a previous optout exists:

        Keep it but change date and origin"""
        mail = MailFactory(message=self.message)
        first_optout = OptOutFactory(
            identifier=mail.identifier,
            address=mail.recipient,
            origin=OptOut.BY_MAIL)
        first_optout_date = first_optout.creation_date

        response = self.client.post(
            '/h/subscriptions/{}/optout/'.format(mail.identifier))
        self.assertEqual(response.status_code, 302)

        o = OptOut.objects.get(identifier=mail.identifier)
        self.assertEqual(o, first_optout)
        self.assertEqual(o.origin, OptOut.BY_WEB)
        self.assertNotEqual(o.creation_date, first_optout_date)

    def test_new_optout_external_post(self):
        self.message.external_optout = True
        self.message.save()
        mail = MailFactory(message=self.message)

        self.assertFalse(OptOut.objects.filter(
            identifier=mail.identifier).exists())
        response = self.client.post(
            '/h/subscriptions/{}/optout/'.format(mail.identifier))
        self.assertEqual(response.status_code, 405)

        self.assertFalse(OptOut.objects.filter(
            identifier=mail.identifier).exists())

    def test_new_optout_external_get(self):
        self.message.external_optout = True
        self.message.save()
        mail = MailFactory(message=self.message)

        self.assertFalse(OptOut.objects.filter(
            identifier=mail.identifier).exists())
        response = self.client.get(
            '/h/subscriptions/{}/optout/'.format(mail.identifier))
        self.assertEqual(response.status_code, 200)
        self.assertIn(
            'should get in touch with him', response.content.decode())

    def test_optout_page_not_preview(self):
        """ Unsubscribe page displays no warning"""
        mail = MailFactory(message=self.message)
        response = self.client.get(
            '/h/subscriptions/{}/optout/'.format(mail.identifier))
        self.assertEqual(response.status_code, 200)

        self.assertNotIn(
            "from a preview message", response.content.decode())

    def test_optout_page_preview(self):
        """ Unsubscribe page displays a warning"""

        pm = PreviewMailFactory(message=self.message)
        response = self.client.get(
            '/h/subscriptions/{}/optout/'.format(pm.identifier))
        self.assertEqual(response.status_code, 200)
        self.assertIn(
            "from a preview message", response.content.decode())



import re
import email
import datetime

import pytz


def extract_domain(address):
    return address.strip().split('@')[-1]


def parse_email_date(date_str):
    """Parse a RFC2822 date

    It is the missing email.utils.parsedate_tz_to_datetime from python stdlib.

    :param date_str: a date and time, from an email header
    :type date_str: str
    :return: a datetime, with tz info on UTC
    :rtype datetime:
    """
    date_tupple = email.utils.parsedate_tz(date_str)
    date = datetime.datetime.fromtimestamp(email.utils.mktime_tz(
        date_tupple), pytz.utc)
    return date


class UniqueEmailAddressParser:
    """ Handle unique addresses like <prefix><base64-uuid>@<domain>

    Child classes must define "domain" and "prefix" attributes
    """

    def __init__(self, prefix, domain):
        self.prefix = prefix
        self.domain = domain

    # Those tricks to accept callable are here to ease settings overriding in
    # unit tests.

    def _get_domain(self):
        if hasattr(self.domain, '__call__'):
            return self.domain()
        return self.domain

    def _get_prefix(self):
        if hasattr(self.prefix, '__call__'):
            return self.prefix()
        return self.prefix

    def _get_regexp(self):
        return '(?P<prefix>{})(?P<uuid>[\w\d\-_]{{22,25}})@{}'.format(
            self._get_prefix(), self._get_domain())

    def new(self, b64uuid=None):
        """
        :param domain: the mail domain
        :param b64uuid: a base64url-encoded uuid, if omited, a new one is
          picked
        """
        from . import mk_base64_uuid

        if not b64uuid:
            b64uuid = mk_base64_uuid()
        return '{}{}@{}'.format(
            self._get_prefix(), b64uuid, self._get_domain())

    def is_valid(self, address):
        address = address or ''
        m = re.match(self._get_regexp(), address)
        return m is not None

    def get_uuid(self, address):
        address = address or ''
        m = re.match(self._get_regexp(), address)
        try:
            return m.group('uuid')
        except AttributeError:
            raise ValueError('Not a valid unique address')



import sys

from celery.signals import celeryd_after_setup

from munch.core.utils import get_worker_type
from munch.core.celery import catch_exception
from munch.core.celery import munch_tasks_router


def register_tasks():
    tasks_map = {
        'core': [
            'munch.apps.domains.tasks.run_domains_validation',
            'munch.apps.domains.tasks.validate_sending_domain_field'
        ],
    }
    munch_tasks_router.import_tasks_map(tasks_map, 'domains')


@celeryd_after_setup.connect
@catch_exception
def configure_worker(instance, **kwargs):
    if get_worker_type() in ['core', 'all']:
        from .tasks import run_domains_validation  # noqa
        from .tasks import validate_sending_domain_field  # noqa
        sys.stdout.write('[domains-app] Registering worker as CORE...')
        munch_tasks_router.register_as_worker('core')



import sys
import socket
import logging
import hashlib
from datetime import timedelta

from django.db import models
from django.db.models import Count
from django.db.models import Prefetch
from django.conf import settings
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

from ..utils.managers import MedianQuerySetMixin
from .utils import UniqueEmailAddressParser
from .validators import rfc3463_regex
from .validators import rfc3463_regex_validator

log = logging.getLogger('munch')

return_path_parser = UniqueEmailAddressParser(
    domain=lambda: settings.RETURNPATH_DOMAIN, prefix='return-')

unsubscribe_parser = UniqueEmailAddressParser(
    domain=lambda: settings.RETURNPATH_DOMAIN, prefix='unsubscribe-')


class TrackingSummary:
    def _get_first_open(self):
        from munch.apps.tracking.models import TrackRecord

        if self._first_open_record is None:
            self._first_open_record = False
            record = TrackRecord.get_from_cache(
                self.mail.identifier, kind='read')
            if record:
                self._first_open_record = record[0]
        return self._first_open_record

    def first_open(self):
        record = self._get_first_open()
        if record:
            return record['creation_date']
        return None

    def open_time(self):
        record = self._get_first_open()
        if record:
            return int(record['properties'].get('reaction_time', 0))
        return 0

    def clicks(self):
        from munch.apps.tracking.models import TrackRecord

        results = []
        records = TrackRecord.get_from_cache(
            self.mail.identifier, kind='click')
        identifiers = [l['properties'].get('link') for l in records]
        links = self.mail.get_original_links(identifiers)
        for record in records:
            results.append(
                {links[record['properties'].get(
                    'link')]: record['creation_date']})
        return results

    def __init__(self, mail):
        self.mail = mail
        self._first_open_record = None


class BaseMailStatusManager(models.Manager):
    def re_run_signals(self, since, mail_path='mail'):
        from munch.core.mail import backend

        limit = 1000
        queryset = self.filter(
            creation_date__gte=timezone.now() - timedelta(seconds=since)).only(
                'pk', 'source_ip', 'destination_domain',
                'status', 'creation_date', mail_path).prefetch_related(
                    Prefetch(
                        mail_path,
                        self.model.mail.field.remote_field
                        .model.objects.all().only('identifier')))
        count = queryset.count()
        rest = count % limit
        for page in range(0, int(count / limit)):
            start = page * limit
            end = start + limit
            sys.stdout.flush()
            sys.stdout.write('Re running from {} to {} (total:{})...\r'.format(
                start, end, count))
            for mailstatus in queryset[start:end]:
                mailstatus.pk = None
                backend.pre_save_mailstatus_signal(
                    sender=self, instance=mailstatus, raw=False)
                backend.post_save_mailstatus_signal(
                    sender=self, instance=mailstatus, created=True, raw=False)

        if rest:
            sys.stdout.write(
                'Re running the latest statuses (total:{})...\r'.format(rest))
        for mailstatus in queryset.all()[count - rest:count]:
            mailstatus.pk = None
            backend.pre_save_mailstatus_signal(
                sender=self, instance=mailstatus, raw=False)
            backend.post_save_mailstatus_signal(
                sender=self, instance=mailstatus, created=True, raw=False)

        return count


class BaseMailStatusQuerySet(models.QuerySet):
    def bounces(self):
        return self.filter(status__in=(
            AbstractMailStatus.BOUNCED, AbstractMailStatus.DROPPED))

    def duration(self):
        """ Duration on which spans a qs of MailStatus

        :returns: a datetime.timedelta
        """
        extrems = self.aggregate(
            end=models.Max('creation_date'), start=models.Min('creation_date'))

        if not extrems['end'] or not extrems['start']:
            # case of empty qs (None values) -> zero-timedelta
            return timedelta()
        else:
            return extrems['end'] - extrems['start']

    def refresh_mail_cache(self):
        """
        Refresh Mail attributes depending on mailstatuses

        Some attributes on Mail model are bare cache of linked Mailstatuses,
        that function goes through all the mailstatus and syncs their values to
        coresponding Mail instances.
        """
        for status in self.order_by('creation_date'):
            AbstractMailStatus.update_mail_status(
                AbstractMailStatus, status, raw=True)


class AbstractMailStatus(models.Model):
    DELIVERED = 'delivered'
    BOUNCED = 'bounced'
    DELAYED = 'delayed'
    DROPPED = 'dropped'
    DELETED = 'deleted'
    UNKNOWN = 'unknown'
    IGNORED = 'ignored'
    QUEUED = 'queued'
    SENDING = 'sending'

    STATUS_CHOICES = (
        (UNKNOWN, _("unknown")),
        (IGNORED, _("ignored")),
        (DELETED, _("deleted")),
        (QUEUED, _("queued")),
        (SENDING, _("sending")),
        (DELAYED, _("delayed (retrying)")),
        (DELIVERED, _("delivered to remote MTA")),
        (DROPPED, _("not delivered (soft-bounce)")),
        (BOUNCED, _("not delivered (hard-bounce)")),
    )

    FINAL_STATES = (BOUNCED, DELIVERED, IGNORED, DROPPED)
    STATES = tuple([UNKNOWN, SENDING] + list(FINAL_STATES) + [QUEUED])

    status = models.CharField(
        max_length=15, default=QUEUED,
        choices=STATUS_CHOICES, verbose_name=_('status'))
    creation_date = models.DateTimeField(verbose_name=_('creation_date'))
    source_ip = models.GenericIPAddressField(
        null=True, blank=True, verbose_name=_('source ip'))
    source_hostname = models.CharField(
        max_length=100, null=True,
        blank=True, verbose_name=_('source hostname'))
    destination_domain = models.CharField(
        max_length=150, verbose_name=_('destination domain'))
    status_code = models.CharField(
        max_length=25, blank=True,
        validators=[rfc3463_regex_validator],
        verbose_name=_('SMTP status code'),
        help_text='X.XXX.XXX, as defined in RFC3463')
    raw_msg = models.TextField(verbose_name=_('raw message'))

    objects = BaseMailStatusManager()

    class Meta:
        abstract = True
        get_latest_by = 'creation_date'
        verbose_name_plural = "Mail Statuses"

    def update_mail_status(self):
        try:
            self.mail.curstatus = self.status
            if not self.pk and not self.creation_date:
                self.creation_date = timezone.now()

            if not self.mail.first_status_date:
                self.mail.first_status_date = self.creation_date
            self.mail.latest_status_date = self.creation_date
            self.mail.delivery_duration = self.mail.latest_status_date \
                - self.mail.first_status_date
            # a mail had issues if it had been softbounced at a certain time.
            self.mail.had_delay = self.mail.had_delay or (
                self.status == AbstractMailStatus.DROPPED)
            # warning : save_base_is not part of the public Django API
            models.Model.save_base(self.mail)
        except self.mail.DoesNotExist:
            pass

    def should_optout(self, create=False):
        from munch.apps.optouts.models import OptOut
        from munch.apps.campaigns.models import (
            MailStatus as CampaignsMailStatus)
        from munch.apps.transactional.models import (
            MailStatus as TransactionalMailStatus)

        filter_kwargs = {
            'mail__recipient': self.mail.recipient,
            'status__in': [self.DROPPED, self.BOUNCED]}

        if OptOut.objects.filter(
                address=self.mail.recipient, origin=OptOut.BY_BOUNCE).exists():
            return True

        bounces = list(CampaignsMailStatus.objects.filter(
            **filter_kwargs).only('creation_date'))
        bounces += list(TransactionalMailStatus.objects.filter(
            **filter_kwargs).only('creation_date'))
        relevant_rule = None
        # Categorize and count the bounces in policies
        for rule in settings.CAMPAIGNS['BOUNCE_POLICY']:
            if self.match_policy(rule):
                relevant_rule = rule
                break

        if relevant_rule is None:
            raise ValueError((
                '"{}" do no match any BOUNCE_POLICY '
                'entry, fix settings.').format(self.status_code))

        # For this rule, count the total number of matching bounces
        count = len([b for b in bounces if b.match_policy(relevant_rule)])

        max_bounces = relevant_rule[1]

        if count >= max_bounces:
            if create:
                log.info(
                    'Unsubscribing {} via bounce'.format(self.mail.recipient))
                OptOut.objects.create_or_update(
                    identifier=self.mail.identifier,
                    address=self.mail.recipient,
                    origin=OptOut.BY_BOUNCE)
            return True
        return False

    def match_policy(self, policy):
        """ does it match the bounce handling policy ?

        @param policy a bounce policy: a 3-uplet:
               - a list of matchs (matches beggining of bounce code)
               - how many bounces before optout (int)
               - how many days back are the bounces counted (int)
        @return a boolean
        """
        matches, how_many, days = policy
        for lookup in matches:
            if (self.status_code.startswith(lookup) and
                    (self.creation_date > (
                        timezone.now() - timedelta(days=days)))):
                return True

        return False

    def save(self, *args, **kwargs):
        created = not self.pk

        self.update_mail_status()

        if self.status_code:
            status_code = rfc3463_regex.match(self.status_code)
            if status_code:
                self.status_code = status_code.group(0)

        if not self.pk:
            if not self.creation_date:
                self.creation_date = timezone.now()
            if not self.source_hostname:
                self.source_hostname = socket.getfqdn()
        super().save(*args, **kwargs)

        if created and self.status in [self.DROPPED, self.BOUNCED]:
            self.should_optout(create=True)


class BaseMailQuerySet(MedianQuerySetMixin, models.QuerySet):
    def done(self):
        return self.filter(
            curstatus__in=AbstractMailStatus.FINAL_STATES).distinct()

    def pending(self):
        return self.exclude(
            curstatus__in=AbstractMailStatus.FINAL_STATES).distinct()

    def duplicate_addrs(self):
        """ Returns a list of couples (count, address) for duplicate mail address

        Usefull on a restricted queryset, for example to find duplicate
        recipients.
        """
        return self.values_list(
            'recipient').annotate(Count('recipient')).exclude(
                recipient__count=1)

    def opened(self):
        """ With triggered open tracker """
        from munch.apps.tracking.models import TrackRecord

        identifiers = self.distinct().values_list('identifier', flat=True)
        opened_identifiers = TrackRecord.objects.filter(
            identifier__in=identifiers, kind='read').values_list(
                'identifier', flat=True)
        return self.filter(identifier__in=opened_identifiers).distinct()

    def clicked(self):
        """ With at least one triggered click tracker """
        from munch.apps.tracking.models import TrackRecord

        identifiers = self.distinct().values_list('identifier', flat=True)
        clicked_identifiers = TrackRecord.objects.filter(
            identifier__in=identifiers, kind='click').values_list(
                'identifier', flat=True)
        return self.filter(identifier__in=clicked_identifiers).distinct()

    def duration(self):
        """ Duration on which spans a qs of MailStatus

        :returns: a timedelta
        """
        extrems = self.aggregate(
            end=models.Max('latest_status_date'),
            start=models.Min('first_status_date'))

        if not extrems['end'] or not extrems['start']:
            # case of empty qs (None values) -> zero-timedelta
            return timedelta()
        return extrems['end'] - extrems['start']

    def last_status_counts(self):
        """ Returns all existing curstatus values in qs """
        statuses = {k: v for k, v in self.values_list(
            'curstatus').annotate(Count('curstatus'))}

        for status in AbstractMailStatus.STATES:
            if status not in statuses:
                statuses[status] = 0
        return statuses

    def info_counts(self):
        """ Counts on certain usefull characteristics of Mails """
        return {
            'done': self.done().count(),
            'had_delay': self.filter(had_delay=True).count(),
            'in_transit': self.pending().count(),
            'total': self.count()}

    def update_cached_fields(self):
        """ Updated cached fields

        Do not have to be called in normal time, and is costly
        (~2h on dev machine for ~1500k Mails)
        """
        # that is not possible without raw SQL :
        # self.with_bounds().update(
        #     first_status_date=F('start'),
        #     latest_status_date=F('end'),
        #     delivery_duration=(F('statuses_delta') - F('start'))
        # )

        qs = self.prefetch_related('statuses').with_bounds()
        log.debug('Starting')
        for i, mail in enumerate(qs):
            mail.first_status_date = mail.start
            mail.latest_status_date = mail.end
            mail.delivery_duration = mail.end - mail.start
            statuses = list(mail.statuses.all())
            # doing the sort on SQL level would ruin prefetch_related benefit
            statuses.sort(key=lambda x: x.creation_date)
            if len(statuses) > 0:
                mail.curstatus = statuses[-1].status
            mail.save()
            if (i % 1000) == 0:
                log.debug(i)


class AbstractMail(models.Model):
    recipient = models.EmailField(verbose_name=_('recipient'))
    creation_date = models.DateTimeField(
        default=timezone.now, verbose_name=_('creation date'))
    first_status_date = models.DateTimeField(
        blank=True, null=True, verbose_name=_('first status date'))
    latest_status_date = models.DateTimeField(
        blank=True, null=True, verbose_name=_('latest status date'))
    delivery_duration = models.DurationField(
        null=True, verbose_name=_('delivery duration'))
    had_delay = models.BooleanField(default=False, verbose_name=_('had delay'))

    objects = BaseMailQuerySet.as_manager()

    class Meta:
        abstract = True
        verbose_name = _('mail')
        verbose_name_plural = _('mails')

    def __str__(self):
        return self.identifier

    @property
    def unsubscribe_addr(self):
        return unsubscribe_parser.new(self.identifier)

    @property
    def envelope_from(self):
        return return_path_parser.new(self.identifier)

    def tracking_info(self):
        return TrackingSummary(self)

    def get_category(self):
        if hasattr(self, 'batch'):
            return self.batch.category
        if hasattr(self, 'message'):
            return self.message.category


class RawMailManager(models.Manager):
    def get_or_create(self, defaults=None, **kwargs):
        content = kwargs.get('content')
        hasher = hashlib.md5()
        if isinstance(content, bytes):
            hasher.update(content)
        else:
            hasher.update(content.encode('utf-8'))
        signature = hasher.hexdigest()
        try:
            return self.get(signature=signature), False
        except self.model.DoesNotExist:
            return self.create(
                signature=signature, content=content), True


class RawMail(models.Model):
    signature = models.CharField(max_length=32, verbose_name=_('signature'))
    content = models.TextField(verbose_name=_('content'))

    objects = RawMailManager()



from datetime import datetime
from email.utils import format_datetime

import pytz
import factory.django
from faker import Factory as FakerFactory

from munch.core.mail.models import RawMail
from munch.apps.users.tests.factories import UserFactory

from ..models import Mail
from ..models import get_mail_identifier

faker = FakerFactory.create()


class MailFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Mail

    identifier = factory.LazyAttribute(lambda x: get_mail_identifier())
    creation_date = factory.LazyAttribute(lambda x: datetime.now(pytz.UTC))
    headers = factory.LazyAttribute(lambda x: {
        'To': faker.email(), 'Date': format_datetime(datetime.now(pytz.UTC))})
    message = factory.LazyAttribute(
        lambda x: RawMail.objects.create(content='Test'))
    author = factory.SubFactory(UserFactory)



from django.db.models.signals import pre_save
from django.db.models.signals import post_save

from munch.core.mail import backend

from .models import MailStatus

# MailStatus
pre_save.connect(backend.pre_save_mailstatus_signal, sender=MailStatus)
post_save.connect(backend.post_save_mailstatus_signal, sender=MailStatus)



import sys
import logging

from django.conf import settings
from django.utils.module_loading import import_string
from slimta.queue import QueueError
from slimta.relay.blackhole import BlackholeRelay


from munch.core.mail.backend import Backend

log = logging.getLogger(__name__)


class QueueStartupFatalError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return 'FATAL, abording startup: {}'.format(self.msg)


class TransactionalQueue:
    def __init__(self, relay):
        # Initialize default queue policies
        self.queue_policies = []
        for path in settings.TRANSACTIONAL.get('QUEUE_POLICIES', []):
            try:
                policy = import_string(path)
            except ImportError:
                raise QueueStartupFatalError(
                    '{} points to inexistant queue policy'.format(path))
            except TypeError:
                raise QueueStartupFatalError(
                    '{} is not a valid QueuePolicy'.format(path))
            self.queue_policies.append(policy())

    def kill(self):
        pass

    def enqueue(self, envelope):
        try:
            envelopes = self._run_policies(envelope)
        except QueueError as exc:
            return [(envelope, exc)]
        ids = [self._initiate_attempt(env) for env in envelopes]
        results = list(zip(envelopes, ids))
        return results

    def _run_policies(self, envelope):
        results = [envelope]

        def recurse(current, i):
            try:
                policy = self.queue_policies[i]
            except IndexError:
                return
            ret = policy.apply(current)
            if ret:
                results.remove(current)
                results.extend(ret)
                for env in ret:
                    recurse(env, i + 1)
            else:
                recurse(current, i + 1)

        recurse(envelope, 0)
        return results

    def _initiate_attempt(self, envelope, attempts=0):
        """ First method called for a sending attempt """
        backend = Backend(
            build_envelope_task_path=(
                'munch.apps.transactional.utils.get_envelope'),
            mailstatus_class_path=(
                'munch.apps.transactional.models.MailStatus'),
            record_status_task_path=(
                'munch.apps.transactional.status.record_status'))
        return backend.send_envelope(envelope, attempts=attempts)


def setup_queue():
    # We only need a dummy slimta relay here
    # only to allow for queue creation
    relay = BlackholeRelay()

    # Set up celery queue
    queue = TransactionalQueue(relay)

    return relay, queue


try:
    # This has to be done, on import, so that celery tasks get set up and
    # available to workers
    relay, queue = setup_queue()
except QueueStartupFatalError:
    log.error("Error while starting Queue", exc_info=True)
    sys.exit(1)



from django import forms

from .models import Contact
from .models import CollectedContact


class ParentSubscriptionForm(forms.ModelForm):
    class Meta:
        fields = ['address']


class ContactSubscriptionForm(ParentSubscriptionForm):
    class Meta(ParentSubscriptionForm.Meta):
        model = Contact


class CollectedContactSubscriptionForm(ParentSubscriptionForm):
    class Meta(ParentSubscriptionForm.Meta):
        model = CollectedContact



from django.conf import settings

from munch.core.mail.utils.munchers import HeadersMuncherRunner


# Filters definition
headers_filters = HeadersMuncherRunner(
    settings.TRANSACTIONAL, 'HEADERS_FILTERS')



from collections import OrderedDict

from django.utils import six
from rest_framework.relations import HyperlinkedRelatedField

from munch.core.utils.permissions import MunchResourcePermission


class ScopedHyperLinkedRelatedField(HyperlinkedRelatedField):
    """ A related field whose content is limited by a permission

    It is already enforced in validation, the purpose of this class is to offer
    the right list in dropdowns.

    Can optionally override the `permission_class` attribute, it must define a
    filter_queryset() function.
    """
    def get_permission(self, model_class):
        permission_class = getattr(self, 'permission_class', None)
        if not permission_class:
            permission_class = MunchResourcePermission.mk_class(model_class)

        return permission_class()

    def get_queryset(self):
        qs = super().get_queryset()
        perm = self.get_permission(qs.model)
        return perm.filter_queryset(qs, self.context.get('request'))

    @property
    def choices(self):
        """
        Copy from parent needed cause @property calls parent
        get_queryset() otherwise (weird isn't it ?)
        """
        queryset = self.get_queryset()
        if queryset is None:
            return {}
        return OrderedDict([
            (
                six.text_type(self.to_representation(item)),
                six.text_type(item)
            )
            for item in queryset
        ])



import datetime

from django.test import TestCase
from django.utils import timezone
from django.core.exceptions import ValidationError
from libfaketime import fake_time
from faker import Factory as FakerFactory

from munch.apps.users.tests.factories import UserFactory
from munch.apps.optouts.tests.factories import OptOutFactory

from ..models import Mail
from ..models import OptOut
from ..models import Message
from ..models import MailStatus

from .factories import MailFactory
from .factories import MessageFactory
from .factories import MailStatusFactory

faker = FakerFactory.create()


class MailManagerTest(TestCase):
    def setUp(self):
        self.user = UserFactory()

        self.message = MessageFactory(status='message_ok', author=self.user)

        mail = MailFactory(message=self.message)
        MailStatusFactory(mail=mail)
        MailStatusFactory(mail=mail, status='sending')
        MailStatusFactory(mail=mail, status='delivered')

        mail = MailFactory(message=self.message)
        MailStatusFactory(mail=mail)
        MailStatusFactory(mail=mail, status='sending')
        MailStatusFactory(mail=mail, status='bounced')

        mail = MailFactory(message=self.message)
        MailStatusFactory(mail=mail)
        MailStatusFactory(mail=mail, status='sending')
        MailStatusFactory(mail=mail, status='dropped', raw_msg='greylisted')
        MailStatusFactory(mail=mail, status='dropped', raw_msg='greylisted')

    def test_all_legit(self):
        legit_objs = set(self.message.mails.legit_for(self.message))

        all_objs = set(self.message.mails.all())
        self.assertEqual(legit_objs, all_objs)

    def test_legit_some_optout(self):
        mail = self.message.mails.first()
        OptOutFactory(
            author=self.message.author, category=self.message.category,
            identifier=mail.identifier, address=mail.recipient)

        legit_objs = set(self.message.mails.legit_for(self.message))
        all_objs = set(self.message.mails.all())

        self.assertIn(mail, all_objs)
        self.assertNotIn(mail, legit_objs)

    def test_legit_some_optout_override(self):
        mail = self.message.mails.first()
        OptOutFactory(
            author=self.message.author, category=self.message.category,
            identifier=mail.identifier, address=mail.recipient)

        legit_objs = set(self.message.mails.legit_for(
            self.message, include_optouts=True))
        all_objs = set(self.message.mails.all())

        self.assertEqual(legit_objs, all_objs)

    def test_legit_some_bounce_optout_override(self):
        # Change the optout type to bounce
        mail = self.message.mails.filter(curstatus='delivered').first()
        opt_out = OptOutFactory(
            identifier=mail.identifier,
            address=mail.recipient,
            origin=OptOut.BY_BOUNCE)

        # add a mail coresponding to an optout on a previous message
        # jane has two reasons not to receive the mail :
        # bounce-optout+regular-optout
        message = MessageFactory(author=self.user)
        MailFactory(recipient=mail.recipient, message=message)
        legit_objs = set(self.message.mails.legit_for(
            self.message, include_optouts=True, include_bounces=True))
        all_objs = set(self.message.mails.all())

        self.assertEqual(legit_objs, all_objs)

    def test_legit_some_bounce_optout(self):
        # Change the optout type to bounce
        mail = self.message.mails.filter(curstatus='bounced').first()
        opt_out = OptOutFactory(
            identifier=mail.identifier,
            address=mail.recipient,
            origin=OptOut.BY_BOUNCE)

        message = MessageFactory(author=self.user)

        # add a mail coresponding to an optout on a previous message
        # jane has two reasons not to receive the mail :
        # bounce-optout+regular-optout
        MailFactory(recipient=mail.recipient, message=message)
        legit_objs = set(self.message.mails.legit_for(
            self.message, include_optouts=True))
        all_objs = set(self.message.mails.all())

        self.assertNotEqual(legit_objs, all_objs)

    def test_done(self):
        self.assertEqual(Mail.objects.done().count(), 3)

    def test_pending(self):
        message = MessageFactory(status='message_ok', author=self.user)

        mail_1 = MailFactory(message=message)
        MailStatusFactory(mail=mail_1)
        MailStatusFactory(mail=mail_1, status=MailStatus.SENDING)

        mail_2 = MailFactory(message=message)
        MailStatusFactory(mail=mail_2)
        MailStatusFactory(mail=mail_2, status=MailStatus.SENDING)

        mail_3 = MailFactory(message=message)
        MailStatusFactory(mail=mail_3)
        MailStatusFactory(mail=mail_3, status=MailStatus.SENDING)
        MailStatusFactory(mail=mail_3, status=MailStatus.DELIVERED)

        self.assertEqual(Mail.objects.filter(
            pk__in=[mail_1.pk, mail_2.pk, mail_3.pk]).pending().count(), 2)

    def test_with_bounds(self):
        Mail.objects.all().delete()
        Message.objects.all().delete()
        MailStatus.objects.all().delete()

        with fake_time('2016-10-10 08:00:00'):
            message = MessageFactory(status='message_ok', author=self.user)
            mail_1 = MailFactory(message=message)
            mail_2 = MailFactory(message=message)
            mail_3 = MailFactory(message=message)

        with fake_time('2016-10-10 08:00:05'):
            MailStatusFactory(mail=mail_1)
            MailStatusFactory(mail=mail_2)
            MailStatusFactory(mail=mail_3)
        with fake_time('2016-10-10 08:00:07'):
            MailStatusFactory(mail=mail_1, status=MailStatus.SENDING)
            MailStatusFactory(mail=mail_2, status=MailStatus.SENDING)
            MailStatusFactory(mail=mail_3, status=MailStatus.SENDING)
        with fake_time('2016-10-10 08:00:57'):
            MailStatusFactory(mail=mail_1, status=MailStatus.DELIVERED)
            MailStatusFactory(mail=mail_2, status=MailStatus.BOUNCED)
            MailStatusFactory(mail=mail_3, status=MailStatus.DROPPED)

        qs = Mail.objects.with_bounds().filter(
            curstatus=MailStatus.BOUNCED).first()
        with fake_time('2016-10-10 08:00:00'):
            self.assertEqual(qs.start, timezone.now())

        with fake_time('2016-10-10 08:00:57'):
            self.assertEqual(qs.end, timezone.now())

        self.assertEqual(qs.delivery_duration.total_seconds(), 57.0)

    def test_median(self):
        with fake_time('2016-10-10 08:00:00'):
            mail_1 = MailFactory(message=self.message)
            mail_2 = MailFactory(message=self.message)
            MailStatusFactory(mail=mail_1)
            MailStatusFactory(mail=mail_2)
        with fake_time('2016-10-10 08:00:05'):
            MailStatusFactory(mail=mail_1, status='sending')
            MailStatusFactory(mail=mail_2, status='sending')
        with fake_time('2016-10-10 08:00:30'):
            MailStatusFactory(mail=mail_1, status='delivered')
        with fake_time('2016-10-10 08:01:00'):
            MailStatusFactory(mail=mail_2, status='delivered')
        median = Mail.objects.filter(
            pk__in=[mail_1.pk, mail_2.pk]).median('delivery_duration')
        self.assertEqual(median, datetime.timedelta(seconds=45))

    def test_median_zeroitem(self):
        median = Mail.objects.none().median('delivery_duration')
        self.assertEqual(median, datetime.timedelta())

    def test_had_delay(self):
        now = timezone.now()

        mail_1 = MailFactory(message=self.message)
        mail_2 = MailFactory(message=self.message)
        self.message.status = Message.SENDING
        self.message.save()

        MailStatusFactory(
            mail=mail_1, status=MailStatus.DELIVERED,
            creation_date=now + datetime.timedelta(minutes=5))
        MailStatusFactory(
            mail=mail_2, status=MailStatus.DROPPED,
            creation_date=now + datetime.timedelta(minutes=5))

        mail_1 = Mail.objects.get(pk=mail_1.pk)
        mail_2 = Mail.objects.get(pk=mail_2.pk)
        self.assertEqual(mail_1.had_delay, False)
        self.assertEqual(mail_2.had_delay, True)

    def test_update_cached_fields(self):
        with fake_time('2016-10-10 08:00:00'):
            mail = MailFactory(
                recipient=faker.email(), message=self.message)
            now = timezone.now()
            d = now
            for status in [
                    MailStatus.QUEUED, MailStatus.SENDING, MailStatus.DROPPED]:
                MailStatusFactory(
                    mail=mail, status=status, creation_date=d, raw_msg='test1')
                d = now + datetime.timedelta(minutes=5)

        Mail.objects.update(
            latest_status_date=None, first_status_date=None,
            delivery_duration=None, curstatus=MailStatus.SENDING)
        Mail.objects.filter(message=self.message).update_cached_fields()

        mail = Mail.objects.get(pk=mail.pk)
        self.assertEqual(mail.delivery_duration.total_seconds(), 300)
        self.assertEqual(
            mail.latest_status_date, now + datetime.timedelta(minutes=5))
        self.assertEqual(mail.first_status_date, now)
        self.assertEqual(mail.curstatus, MailStatus.DROPPED)
        self.assertEqual(mail.had_delay, True)

    def test_mail_with_source(self):
        self.message.mails.all().delete()

        MailFactory(
            message=self.message, source_type='zim-bra', source_ref='list-42')

    def test_mail_with_properties(self):
        self.message.mails.all().delete()

        mail_1 = MailFactory(message=self.message, properties={'gender': 'm'})
        MailFactory(message=self.message, properties={'gender': 'f'})

        filtered = Mail.objects.filter(properties__gender='m')
        self.assertEqual(filtered.count(), 1)
        self.assertEqual(filtered.first().recipient, mail_1.recipient)

    def test_mail_with_bad_source_format(self):
        self.message.mails.all().delete()

        with self.assertRaises(ValidationError):
            m = MailFactory(
                message=self.message,
                source_type='zim bra',
                source_ref='list-42')
            m.full_clean()



import uuid
from urllib.parse import urljoin

from django.db import models
from django.db.models import Count
from django.urls import reverse
from django.conf import settings
from django.contrib.postgres import fields
from django.contrib.postgres.fields import JSONField

from munch.core.utils.models import AbstractOwnedModel
from munch.apps.users.models import MunchUser
from munch.apps.campaigns.validators import slug_regex_validator

from . import backends
from .validators import properties_schema_validator

PROP_TYPES = ('Char', 'Integer', 'Date', 'DateTime', 'Boolean', 'Float')


class ContactListPolicy(models.Model):
    name = models.CharField('nom', max_length=100)
    description = models.TextField('description')

    def get_backend(self):
        return getattr(backends, self.name)

    def __str__(self):
        return self.name

    def apply(self, item, policies_list):
        """
        Applies the current policy to the item (basically, a Contact).
        The backend could take knowledge of the full policies list, in order to
        adapt its behaviour.
        """
        self.get_backend()().apply(item, policies_list)

    class Meta:
        verbose_name = 'politique de file'
        verbose_name_plural = 'politiques de files'


class ContactQueuePolicyAttribution(models.Model):
    policy = models.ForeignKey(ContactListPolicy)
    contact_queue = models.ForeignKey('contacts.ContactQueue')

    class Meta:
        unique_together = (('policy', 'contact_queue', ), )


class ContactListPolicyAttribution(models.Model):
    policy = models.ForeignKey(ContactListPolicy)
    contact_list = models.ForeignKey('contacts.ContactList')

    class Meta:
        unique_together = (('policy', 'contact_list', ), )


class AbstractContact(AbstractOwnedModel):
    PENDING = 'pending'
    BOUNCED = 'bounced'
    EXPIRED = 'expired'
    OK = 'ok'
    CONSUMED = 'consumed'

    properties = fields.HStoreField('propriétés', default={})
    address = models.EmailField('adresse e-mail')
    creation_date = models.DateTimeField('date de création', auto_now_add=True)
    update_date = models.DateTimeField('date de modification', auto_now=True)
    uuid = models.UUIDField(editable=False, default=uuid.uuid4)
    status = models.CharField(
        'statut', max_length=50, choices=(
            (PENDING, 'en attente'),
            (BOUNCED, 'échec bounce'),
            (EXPIRED, 'expiré'),
            (OK, 'ok'),
            (CONSUMED, 'consommé')), default=PENDING)
    subscription_ip = models.GenericIPAddressField(
        'adresse IP d’envoi', default='127.0.0.1')

    class Meta(AbstractOwnedModel.Meta):
        abstract = True
        verbose_name = 'contact'

    def apply_policies(self):
        """
        Executes each policy registered with the ContactList.
        """
        if getattr(self, self.contact_list_path).policies.count() == 0:
            self.status = self.OK
            self.save()
        else:
            policies = getattr(self, self.contact_list_path).policies.all()
            policies_names = [p.name for p in policies]
            for policy in policies:
                policy.apply(self, policies_names)


class Contact(AbstractContact):
    contact_list = models.ForeignKey(
        'ContactList', related_name='contacts')

    owner_path = 'contact_list__author__organization'
    author_path = 'contact_list__author'
    contact_list_path = 'contact_list'

    class Meta:
        unique_together = (('contact_list', 'address'), )

    def __str__(self):
        return self.address


class CollectedContact(AbstractContact):
    contact_queue = models.ForeignKey(
        'ContactQueue', related_name='collected_contacts')

    owner_path = 'contact_queue__author__organization'
    author_path = 'contact_queue__author'
    contact_list_path = 'contact_queue'

    class Meta:
        verbose_name = 'adresse collectée'
        verbose_name_plural = 'adresses collectées'
        unique_together = (('contact_queue', 'address'), )


class AbstractContactList(AbstractOwnedModel):
    author = models.ForeignKey(MunchUser, verbose_name='auteur')
    properties = fields.HStoreField('propriétés', default={})
    source_type = models.CharField(
        blank=True, max_length=100, validators=[slug_regex_validator])
    source_ref = models.CharField(blank=True, max_length=200)
    uuid = models.UUIDField(editable=False, default=uuid.uuid4)

    owner_path = 'author__organization'
    author_path = 'author'

    class Meta:
        abstract = True

    @property
    def subscription(self):
        return urljoin(
            settings.APPLICATION_URL.strip('/'),
            reverse('subscription', kwargs={'uuid': self.uuid}))


class ContactList(AbstractContactList):
    DEFAULT_CONTACT_FIELDS = [
        {'name': 'Prénom', 'type': 'Char', 'required': False},
        {'name': 'Nom', 'type': 'Char', 'required': False}]
    name = models.CharField('nom', max_length=100)
    contact_fields = JSONField(
        'champs des contacts', default=DEFAULT_CONTACT_FIELDS,
        validators=[properties_schema_validator])
    policies = models.ManyToManyField(
        ContactListPolicy,
        verbose_name='politiques de la file',
        through=ContactListPolicyAttribution)

    class Meta:
        verbose_name = 'liste de contacts'
        verbose_name_plural = 'listes de contacts'
        unique_together = (('author', 'name'), )

    def __str__(self):
        return self.name

    def contacts_count(self):
        return self.contacts.count()


class ContactQueue(AbstractContactList):
    DEFAULT_CONTACT_FIELDS = []
    contact_fields = JSONField(
        'champs des contacts', default=DEFAULT_CONTACT_FIELDS,
        validators=[properties_schema_validator])
    policies = models.ManyToManyField(
        ContactListPolicy,
        verbose_name='politiques de la file',
        through=ContactQueuePolicyAttribution)

    def contacts_count(self):
        """Returns a dict with contact count for each status in the queue.
        """
        return {c['status']: c['count'] for c in
                self.collected_contacts.values('status').annotate(
                    count=Count('status'))}

    def consume(self, status=AbstractContact.OK):
        contacts = self.collected_contacts.filter(status=status)
        if status == AbstractContact.OK:
            contacts_list = list(contacts)
            contacts.update(status=AbstractContact.CONSUMED)
            contacts = contacts_list
            # because QuerySet.update() seems to consume the list
        return contacts



import html
import email
import magic
import urllib
import logging
import re
from os.path import join
from base64 import b64encode

import chardet
import html2text
import lxml.html
import lxml.etree
import celery
import django.core.mail.message
from django.conf import settings
from django.db import models
from django.db import transaction
from django.utils import timezone
from django.urls import reverse
from django.core.mail import EmailMultiAlternatives
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _
from django.contrib.postgres.fields import JSONField
from django.contrib.postgres.fields import HStoreField
from django.template.loader import render_to_string
from django_fsm import transition
from slimta.envelope import Envelope

from munch.core.models import Category
from munch.core.mail.backend import Backend
from munch.core.utils import save_timer
from munch.core.utils.models import AbstractOwnedModel
from munch.core.mail.utils import mk_base64_uuid
from munch.core.mail.utils import mk_msgid
from munch.core.mail.utils.emails import NotificationMessage
from munch.core.mail.models import AbstractMail
from munch.core.mail.models import AbstractMailStatus
from munch.core.mail.models import BaseMailStatusManager
from munch.core.models import ValidationSignalsModel
from munch.apps.spamcheck import SpamResult
from munch.apps.spamcheck import SpamChecker
from munch.apps.spamcheck import SpamCheckerError
from munch.apps.domains.models import SendingDomain
from munch.apps.users.models import MunchUser
from munch.apps.optouts.models import OptOut
from munch.apps.tracking.utils import WebKey
from munch.apps.tracking.utils import get_msg_links
from munch.apps.tracking.models import TrackRecord
from munch.apps.tracking.models import READ_BROWSER

from .fields import FSMAutoField
from .validators import slug_regex_validator
from .exceptions import WrongHTML
from .managers import MailManager
from .managers import MailQuerySet

from .managers import MailStatusQuerySet
from .munchers import post_template_html_generation
from .munchers import post_individual_html_generation
from .munchers import post_individual_plaintext_generation
from .munchers import post_headers_generation

log = logging.getLogger('munch')

email.charset.add_charset('utf-8', email.charset.SHORTEST, None, 'utf-8')


class MailStatus(AbstractOwnedModel, AbstractMailStatus):
    mail = models.ForeignKey(
        'Mail', related_name='statuses', verbose_name=_('mail'))

    objects = BaseMailStatusManager.from_queryset(MailStatusQuerySet)()

    class Meta(AbstractOwnedModel.Meta):
        verbose_name = _('email status')
        verbose_name_plural = _('email statuses')
        get_latest_by = 'creation_date'

    owner_path = 'mail__message__author__organization'
    author_path = 'mail__message__author'

    def __str__(self):
        return _('Message {} is {}').format(self.mail.identifier, self.status)

    def get_organization(self):
        return self.mail.message.author.organization

    def update_message_status(self):
        try:
            if self.mail.curstatus in MailStatus.FINAL_STATES:
                # If there is nothing left to send,
                # change the state of the message !
                if self.mail.message.status == Message.SENDING:
                    remaining = self.mail.message.mails.legit_for(
                        self.mail.message).pending().count()
                    if remaining == 0:
                        self.mail.message.status = Message.SENT
                        self.mail.message.save()
        except Mail.DoesNotExist:
            pass

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self.update_message_status()


class TolerantForeignKey(models.ForeignKey):
    """ A ForeignKey which can point to an unsaved object

    It's back to django >= 1.7 behaviour
    to allow mem-only message making
    """
    allow_unsaved_instance_assignment = True


def get_base_mail_identifier():
    return mk_base64_uuid('c-')


class BaseMail(AbstractOwnedModel, AbstractMail):
    identifier = models.CharField(
        max_length=35, db_index=True, unique=True,
        default=get_base_mail_identifier, verbose_name=_('identifier'))
    message = TolerantForeignKey(
        'Message', related_name='%(class)ss', verbose_name=_('message'))
    properties = HStoreField(
        null=True, blank=True, verbose_name=_('properties'))
    source_type = models.CharField(
        blank=True, max_length=100,
        validators=[slug_regex_validator], verbose_name=_('source type'))
    source_ref = models.CharField(
        blank=True, max_length=200, verbose_name=_('source reference'))
    # curstatus is overwritten because we want FSMAutoField here
    curstatus = FSMAutoField(
        default=AbstractMailStatus.UNKNOWN,
        choices=AbstractMailStatus._meta.get_field('status').choices,
        verbose_name=_('current status'))

    class Meta(AbstractOwnedModel.Meta):
        abstract = True

    owner_path = 'message__author__organization'
    author_path = 'message__author'

    def __str__(self):
        return _('Message {} to {}').format(self.identifier, self.recipient)

    def get_organization(self):
        return self.message.author.organization

    def get_category(self):
        if self.message.category:
            return self.message.category

    def get_related_optout(self):
        """ Gets the related OptOut

        That is not an optout on that email, but rather an optout on another
        email to the same recipient.

        :returns: an OptOut object or None
        """
        if self.message.category:
            qs = OptOut.objects.filter(category=self.message.category)
        else:
            qs = OptOut.objects.filter(
                author__organization=self.message.author.organization)

        return qs.for_email(self.recipient)

    @property
    def abuse_url(self):
        relative_url = reverse(
            'abuse-report', kwargs={'identifier': self.identifier})
        return urllib.parse.urljoin(self.message.get_app_url(), relative_url)

    @property
    def unsubscribe_url(self):
        relative_url = reverse(
            'unsubscribe', kwargs={'identifier': self.identifier})
        return urllib.parse.urljoin(self.message.get_app_url(), relative_url)

    @property
    def web_view_url(self):
        relative_url = reverse(
            'message_web_view', kwargs={'identifier': self.message.identifier})

        abs_url = urllib.parse.urljoin(
            self.message.get_app_url(), relative_url)

        if self.message.track_open:
            return abs_url + '?web_key={}'.format(
                WebKey.from_instance(self.identifier).token)
        else:
            return abs_url

    def as_message(self, with_body=True):
        return self.message.to_mail(self, with_body=with_body)

    def get_headers(self):
        message = self.as_message(with_body=False).message()
        return dict(message)

    def as_envelope(self):
        message = self.as_message()
        envelope = Envelope()
        envelope.parse(message.message().as_bytes())
        envelope.sender = message.from_email
        envelope.recipients.append(self.recipient)
        return envelope

    def get_original_links(self, links):
        original_links = {}
        for link in links:
            original_links[link] = self.message.msg_links.get(link)
        return original_links

    def last_status(self):
        try:
            # Try to use cached version, see api/views.py
            return self.last_status_cached[0]
        except IndexError:
            return None
        except AttributeError:
            try:
                return self.statuses.latest()
            except MailStatus.DoesNotExist:
                return None

    last_status.short_description = _('Last status')

    def is_delivered(self):
        return self.curstatus == 'delivered'

    @transition(field=curstatus,
                source=MailStatus.UNKNOWN, target=MailStatus.QUEUED)
    def celery_enqueued(self):
        pass

    @transition(field=curstatus,
                source=MailStatus.UNKNOWN, target=MailStatus.IGNORED)
    def ignoring_send(self):
        pass

    @transition(field=curstatus,
                source=MailStatus.QUEUED, target=MailStatus.IGNORED)
    def delivery_will_fail(self):
        pass

    @transition(field=curstatus,
                source=MailStatus.QUEUED, target=MailStatus.SENDING)
    def sending_begin(self):
        pass

    @transition(field=curstatus,
                source=MailStatus.SENDING, target=MailStatus.DELAYED)
    def delayed_dsn(self):
        pass

    @transition(field=curstatus,
                source=MailStatus.DELAYED, target=MailStatus.DELIVERED)
    def reception_after_delay(self):
        pass

    @transition(field=curstatus,
                source=MailStatus.DELAYED, target=MailStatus.BOUNCED)
    def hardbounce_after_delay(self):
        pass

    @transition(field=curstatus,
                source=MailStatus.DELAYED, target=MailStatus.DROPPED)
    def softbounce_after_delay(self):
        pass

    @transition(field=curstatus,
                source=MailStatus.DELAYED, target=MailStatus.SENDING)
    def sending_after_delay(self):
        pass

    @transition(field=curstatus,
                source=MailStatus.SENDING, target=MailStatus.DELIVERED)
    def mx_reception(self):
        pass

    @transition(field=curstatus,
                source=MailStatus.SENDING, target=MailStatus.BOUNCED)
    def mx_hardbounce(self):
        pass

    @transition(field=curstatus,
                source=MailStatus.SENDING, target=MailStatus.DROPPED)
    def mx_softbounce(self):
        pass

    @transition(field=curstatus,
                source=MailStatus.DELIVERED, target=MailStatus.BOUNCED)
    def hardbounce_backscatter(self):
        """May happen : local MTA ack, then remote mta backscatters
        """
        pass

    @transition(field=curstatus,
                source=MailStatus.DELIVERED, target=MailStatus.DROPPED)
    def softbounce_backscatter(self):
        """May happen : local MTA ack, then remote mta backscatters
        """
        pass


class Mail(BaseMail):
    objects = MailManager.from_queryset(MailQuerySet)()

    class Meta:
        unique_together = [('recipient', 'message')]


class Message(ValidationSignalsModel, AbstractOwnedModel):
    NEW = 'new'
    MSG_OK = 'message_ok'
    MSG_ISSUES = 'message_issues'
    SENDING = 'sending'
    SENT = 'sent'
    STATUS_CHOICES = (
        (NEW, _('Created')),
        (MSG_OK, _('Message approved')),
        (MSG_ISSUES, _('Message contains issues')),
        (SENDING, _('Sending')),
        (SENT, _('Sent')))

    # Identifier is used as a key for public URLs.
    identifier = models.CharField(
        max_length=35, db_index=True, unique=True, editable=False,
        default=mk_base64_uuid, verbose_name=_('identifier'))

    # User filled
    name = models.CharField(_('name'), max_length=60)

    # Message-related fields (user-filled)
    sender_email = models.EmailField(_('sender address'))
    sender_name = models.CharField(_('sender name'), max_length=50)

    subject = models.CharField(_('subject'), blank=True, max_length=130)
    html = models.TextField(blank=True)

    # Auto-filled at first save
    status = FSMAutoField(_('status'), default=NEW, choices=STATUS_CHOICES)

    author = models.ForeignKey(MunchUser, verbose_name=_('author'))

    creation_date = models.DateTimeField(
        _('creation date'), auto_now_add=True, help_text=_('automatic'))
    send_date = models.DateTimeField(_('send date'), null=True, blank=True)
    completion_date = models.DateTimeField(
        _("complete sent date"), null=True, blank=True)

    # Optional category
    category = models.ForeignKey(
        Category, null=True, blank=True,
        related_name='messages',
        verbose_name=_('category'))

    # Opaque Hash to be used by API clients to store whatever they fancy
    properties = HStoreField(
        verbose_name=_('properties'), null=True, blank=True)

    # Settings
    external_optout = models.BooleanField(
        _('external optout'), default=False, help_text=(
            _(
                "If enabled, this message, instead of optout form, recipients "
                "will be invited to contact you.")))
    detach_images = models.BooleanField(
        _('host images'), default=False,
        help_text=_(
            "Host linked images on our servers instead of original links."))
    track_open = models.BooleanField(
        _('track opens'), default=False,
        help_text=(_("Detect who opens message by adding hidden image.")))
    track_clicks = models.BooleanField(
        _('track clicks'), default=False,
        help_text=_(
            "Detect which links were clicked by replacing "
            "original links by redirections"))

    # Auto-filled (message related)
    spam_score = models.FloatField(
        blank=True, null=True, verbose_name=_('spam score'))
    spam_details = JSONField(
        null=True, blank=True, verbose_name=_('Triggered anti-spam rules'))
    is_spam = models.BooleanField(_('Detected as spam'), default=False)
    spam_check_error = models.TextField(
        null=True, blank=True, verbose_name=_('Spam check error'))
    msg_issue = models.CharField(
        verbose_name=_('message issue'), max_length=300, blank=True, null=True)
    msg_links = HStoreField(
        verbose_name=_('message links'), null=True, blank=True)

    class Meta(AbstractOwnedModel.Meta):
        verbose_name = _('message')
        verbose_name_plural = _('messages')
        default_permissions = AbstractOwnedModel.Meta.default_permissions + (
            'previewsend_organizations', 'previewsend_mine', 'previewsend')

    owner_path = 'author__organization'
    author_path = 'author'

    def __str__(self):
        return self.name

    def get_organization(self):
        return self.author.organization

    def get_sending_domain(self):
        all_domains = SendingDomain.objects
        org_id = self.get_organization()
        qs = all_domains.filter(organization=org_id) | all_domains.filter(
            alt_organizations__in=[org_id])
        organization_domains = qs.distinct()
        return organization_domains.get_from_email_addr(
            self.sender_email, must_raise=False)

    def get_app_url(self):
        from munch.core.mail.utils import get_app_url

        return get_app_url(
            domain=self.get_sending_domain(),
            organization=self.get_organization())

    def willsend_addresses(self):
        return self.mails.legit_for(self).values_list('recipient', flat=True)

    def willnotsend_addresses(self):
        return self.mails.not_legit_for(self).values_list(
            'recipient', flat=True)

    def get_clicks_defaults(self):
        if self.msg_links:
            return {k: 0 for k in list(self.msg_links.values())}
        return {}

    def get_clicks_count(self, click_qs, unique=False):
        clicked = click_qs.count_by_url(
            self.msg_links, include_any=unique, unique=unique)
        links = self.get_clicks_defaults()
        links.update(clicked)
        return links

    @save_timer(name='campaigns.Message.mk_stats')
    def mk_stats(self):
        mail_qs = self.mails.all()
        identifiers = self.mails.all().values_list('identifier', flat=True)
        click_qs = TrackRecord.objects.filter(
            identifier__in=identifiers, kind='click')
        open_qs = TrackRecord.objects.filter(
            identifier__in=identifiers, kind='read').order_by(
            'identifier', 'creation_date')
        browser_views = open_qs.filter(
            properties__source=READ_BROWSER).distinct().count()
        optout_qs = OptOut.objects.filter(
            identifier__in=self.mails.values_list('identifier', flat=True))

        open_time_serie = [int(p[
            'reaction_time']) for p in open_qs.values_list(
            'properties', flat=True) if p['reaction_time']]
        open_median_time = None
        if open_time_serie:
            open_median_time = sum(open_time_serie) / len(open_time_serie)

        return {
            'count': mail_qs.info_counts(),
            'last_status': mail_qs.last_status_counts(),
            'tracking': {
                'opened': mail_qs.opened().count(),
                'open_median_time': open_median_time,
                'clicked': self.get_clicks_count(click_qs, unique=True),
                'clicked_total': self.get_clicks_count(click_qs, unique=False),
                'viewed_in_browser': browser_views
            },
            'timing': {
                'delivery_total': mail_qs.duration().seconds,
                'delivery_median': mail_qs.median('delivery_duration').seconds
            },
            'optout': optout_qs.count_by_origin(with_total=True)
        }

    def to_mail(self, mail, with_body=True):
        """
        @param mail : the Mail instance we want to build the Message for
        @return a EMailMultiAlternatives ready to send
        """

        m = EmailMultiAlternatives(
            subject=self.subject,
            body='',
            from_email=mail.envelope_from,
            to=[mail.recipient])

        m.extra_headers.update({
            settings.CAMPAIGNS['X_USER_ID_HEADER']: str(
                mail.message.author.pk),
            settings.CAMPAIGNS['X_MESSAGE_ID_HEADER']: mail.identifier,
            'Precedence': 'bulk',
            'X-Report-Abuse': 'Please report abuse to {}'.format(
                mail.abuse_url),
            'List-Unsubscribe': '<mailto:{}>, <{}>'.format(
                mail.unsubscribe_addr, mail.unsubscribe_url),
            'Message-ID': mk_msgid(),
            'From': '{} <{}>'.format(
                mail.message.sender_name, mail.message.sender_email),
        })

        post_headers_generation.process(
            m.extra_headers, mail,
            settings.CAMPAIGNS['HEADERS_FILTERS_PARAMS'])

        if with_body:
            plaintext, html = self.mk_body(mail)
            m.body = plaintext
            m.attach_alternative(html, "text/html")

            # attachments
            for a in self.attachments.all():
                m.attach(*a.to_email_attachment())

        return m

    @save_timer(name='campaigns.Message.mk_body')
    def mk_body(self, mail):
        generation_kwargs = {
            'track_open': self.track_open,
            'track_clicks': self.track_clicks,
            'unsubscribe_url': mail.unsubscribe_url,
            'app_url': self.get_app_url(),
            'links_map': self.msg_links or {},
            'mail_properties': mail.properties,
            'mail_identifier': mail.identifier,
            'no_unsubscribe_placehoder_must_raise': True,
            'web_view_url': mail.web_view_url
        }

        plaintext = post_individual_plaintext_generation.process(
            self.mk_plaintext(), **generation_kwargs)

        html = post_individual_html_generation.process(
            self.mk_html(), **generation_kwargs)

        # Use max line length from RFC2822 (78) instead of RFC5322 (998)
        # to force conversion to quoted-printable in almost all cases
        # The idea is to avoid anarchic line-breaks in 7bits
        # formatted mails which cause bad display of some utf-8 characters
        # on some webmails.
        # Note: this might cause problems in non-western character sets...
        django.core.mail.message.RFC5322_EMAIL_LINE_LENGTH_LIMIT = 78
        # Also remove blank lines and leading/trailing spaces in HTML
        # content to make generated body more concise
        for exp in (r'^\s*$', r'^\s+', r'\s+$'):
            html = re.sub(exp, '', html, flags=re.MULTILINE)

        return plaintext, html

    def mk_plaintext(self):
        try:
            h = html2text.HTML2Text()
            h.ignore_images = True
            h.inline_links = False
            h.wrap_links = False
            h.unicode_snob = True  # Prevents accents removing
            h.skip_internal_links = True
            h.ignore_anchors = True
            h.body_width = 0
            h.use_automatic_links = True
            h.ignore_tables = True
        except html.parser.HTMLParseError as e:
            raise WrongHTML(e)

        return h.handle(self.mk_html())

    def mk_html(self):
        """Simply calls configured html template filters

        See settings.CAMPAIGNS['HTML_TEMPLATE_FILTERS']
        """
        # Doctype gets frequently removed by content filters, so we save
        # it...
        doc = lxml.etree.HTML(self.html)
        doctype = ''
        if doc is not None:
            doctype = doc.getroottree().docinfo.doctype

        # ... we process content...
        mangled_content = post_template_html_generation.process(
            self.html,
            detach_images=self.detach_images,
            organization=self.author.organization)

        # And we re-inject it
        return '{}\n{}'.format(doctype, mangled_content)

    def spam_check(self):
        """ Requires the current message to be saved """
        # create dummy message
        mail = Mail(
            recipient='foo@example.com', message=self,
            creation_date=timezone.now(), identifier='i-am-a-test-uuid')
        message = mail.as_message()
        try:
            sc = SpamChecker(
                port=settings.SPAMD_PORT, host=settings.SPAMD_HOST)
        except AttributeError:
            sc = SpamChecker(settings.SPAMD_SERVICE_NAME)

        try:
            return sc.check(message.message().as_string())
        except SpamCheckerError as e:
            error_message = _(
                "Spam checker not available for now: {}").format(e)
            log.error(error_message)
            return SpamResult(-1, False, [], error=error_message)

    def notify(self, state):
        """
        Send email notifications based on current message state

        @param state : a valid value for Message.status
        """
        if not self.author.organization.settings.notify_message_status \
                or not self.author.organization.contact_email:
            return
        if state == Message.SENDING:
            message = NotificationMessage(
                subject=render_to_string(
                    'campaigns/sending_notice_email_subject.txt', {
                        'product_name': settings.PRODUCT_NAME,
                        'name': self.name}).strip(),
                template='campaigns/sending_notice_email.txt',
                render_context={'message': self},
                to=self.author.organization.contact_email,
            )
            message.add_html_part_if_exists(
                'campaigns/sending_notice_email.html')
            message.send()
        elif state == Message.SENT:
            stats = self.mk_stats()
            message = NotificationMessage(
                subject=render_to_string(
                    'campaigns/sending_done_notice_email_subject.txt', {
                        'product_name': settings.PRODUCT_NAME,
                        'name': self.name}).strip(),
                template='campaigns/sending_done_notice_email.txt',
                render_context={'message': self, 'stats': stats},
                to=self.author.organization.contact_email,
            )
            message.add_html_part_if_exists(
                'campaigns/sending_done_notice_email.html')
            message.send()

    def start_sending(self):
        from .tasks import send_mail

        with transaction.atomic():
            # Then, handle all ignored emails (optouts)
            now = timezone.now()
            for i in self.mails.not_legit_for(self):
                MailStatus.objects.create(
                    mail=i, status=MailStatus.IGNORED,
                    creation_date=timezone.now(),
                    raw_msg='Ignored because of previous optout {}'.format(
                        i.get_related_optout()))

            # Then, handle the legit ones
            if not self.send_date:
                self.send_date = timezone.now()
                self.save()
            self.notify(Message.SENDING)
            # locking ?
            legit_mails = self.mails.legit_for(self)
            now = timezone.now()
            MailStatus.objects.bulk_create([
                MailStatus(
                    mail=i, status=MailStatus.QUEUED,
                    creation_date=now, raw_msg='Enqueued in celery')
                for i in legit_mails])
            # bulk_create do not update Mail (too high db cost)
            legit_mails.update(
                curstatus=MailStatus.QUEUED, latest_status_date=now)
            # end_locking ?
            tasks = celery.group([send_mail.s(m.pk) for m in legit_mails])
            log.info('Starting sending {} (#{}) to {} recipients.'.format(
                self, self.pk, len(tasks)))

        tasks.apply_async()

    def has_no_msg_issues(self):
        return not self.has_msg_issues()

    def has_msg_issues(self):
        return self.msg_issue != ''

    def field_changed_and_exists(self, field):
        new_val = getattr(self, field)

        if new_val:
            try:
                old_instance = Message.objects.get(pk=self.pk)
                old_val = getattr(old_instance, field)
                return new_val != old_val
            except Message.DoesNotExist:
                return bool(new_val)

    @save_timer(name='campaigns.Message.build_message')
    def build_message(self):
        if self.html:
            if (self.field_changed_and_exists('html') or
                    self.field_changed_and_exists('subject')):
                if self.status != self.SENT:
                    self.msg_issue = ''
                    try:
                        self.mk_html()
                        self.mk_plaintext()
                    except Exception as e:
                        self.msg_issue = str(e)
                    else:
                        if self.is_spam:
                            self.msg_issue = (
                                _('Message detected as spam'))

                    if self.msg_issue:
                        self.status = self.MSG_ISSUES
                    else:
                        if self.status not in ('sending', 'sent'):
                            self.status = self.MSG_OK
        else:
            self.msg_issue = _('No HTML body given')
            self.status = self.MSG_ISSUES

    @save_timer(name='campaigns.Message.update_spam_score')
    def update_spam_score(self):
        if not settings.CAMPAIGNS['SKIP_SPAM_CHECK'] and (
                self.field_changed_and_exists('html') or
                ((len(self.html) > 0) and
                    self.field_changed_and_exists('subject'))) and (
                        self.status != self.SENDING):

            result = self.spam_check()
            self.spam_score = result.score
            self.spam_details = result.checks
            self.is_spam = result.is_spam
            self.spam_check_error = result.error

    @save_timer(name='campaigns.Message.extract_message_links')
    def extract_message_links(self):
        if self.field_changed_and_exists('html') and self.track_clicks:
            self.msg_links = get_msg_links(self.html)

    @save_timer(name='campaigns.Message.validate_html')
    def validate_html(self):
        errors = []
        if not self.html or self.html.isspace():
            errors.append(_("HTML field can not be empty"))
        elif self.field_changed_and_exists('html'):
            try:
                mail = Mail(
                    recipient='foo@example.com', message=self,
                    creation_date=timezone.now(),
                    identifier='i-am-a-test-uuid')
                self.mk_body(mail)
            except Exception as exc:
                errors.append(str(exc))
        if errors:
            raise ValidationError({'html': errors})

    def clean_fields(self, exclude):
        if html not in exclude:
            self.validate_html()
        super().clean_fields(exclude)

    def clean(self):
        super().clean()
        if self.external_optout and not \
                self.author.organization.can_external_optout:
            raise ValidationError(
                _("External optout forbidden. Please contact us."))

    def save(self, *args, **kwargs):
        self.validate_html()
        self.extract_message_links()
        self.update_spam_score()
        self.build_message()

        # Keep previous object state
        previous_message = None
        if self.pk:
            previous_message = Message.objects.get(pk=self.pk)

        super().save(*args, **kwargs)

        # Check if transition is message_ok to sending
        if previous_message and \
                previous_message.status == Message.MSG_OK and \
                self.status == Message.SENDING:
            self.start_sending()

    ###############
    # Transitions #
    ###############
    @transition(field=status, source=NEW, target=MSG_ISSUES)
    def new_message_issues(self):
        pass

    @transition(field=status, source=NEW, target=MSG_OK)
    def new_message_ok(self):
        pass

    @transition(field=status, source=MSG_OK, target=MSG_ISSUES)
    def updated_message_issues(self):
        pass

    @transition(field=status, source=MSG_ISSUES, target=MSG_OK)
    def updated_message_ok(self):
        pass

    @transition(field=status, source=MSG_OK, target=SENDING)
    def prepare_sending(self):
        if not settings.BYPASS_DNS_CHECKS:
            domain = self.get_sending_domain()
            if not domain or not domain.is_valid():
                raise ValidationError(_(
                    'Your sending domain is not configured correctly. '
                    'Message can not be sent.'))

    @transition(field=status, source=SENDING, target=SENT)
    def complete_sending(self):
        self.completion_date = timezone.now()
        log.info('Finished sending {} (#{}).'.format(self, self.pk))

        self.notify(Message.SENT)


def get_attachment_storage_path(obj, filename):
    return join('attachments', str(obj.message.pk), filename)


class MessageAttachment(AbstractOwnedModel):
    file = models.FileField(
        upload_to=get_attachment_storage_path,
        verbose_name=_('attached file'))
    original_name = models.CharField(
        max_length=100, verbose_name=_('original name'))
    b64size = models.PositiveIntegerField(
        default=0, verbose_name=_('base 64 size'))
    message = models.ForeignKey(
        Message, related_name='attachments', verbose_name=_('message'))

    owner_path = 'message__author__organization'
    author_path = 'message__author'

    class Meta(AbstractOwnedModel.Meta):
        pass

    def save(self, *args, **kwargs):
        self.b64size = len(b64encode(self.file.read()))
        if not self.original_name:
            self.original_name = self.file.name[:100]
        # Go back to the beggining of the file for later
        self.file.seek(0)
        super().save(*args, **kwargs)

    @property
    def size(self):
        # Handle case where blob has been deleted
        try:
            return self.file.size
        except:
            return None

    @property
    def size_in_mail(self):
        return self.b64size

    @property
    def mimetype(self):
        """ Readable mimetype, guessed from file content """
        self.file.open()
        return magic.from_buffer(self.file.read(1024), mime=True)

    @property
    def path(self):
        # Handle case where blob has been deleted
        try:
            return self.file.path
        except:
            return None

    def to_email_attachment(self):
        mime_type = self.mimetype
        self.file.open()
        if mime_type.startswith('text/'):
            charset = chardet.detect(self.file.read(1024))
            self.file.open()
            if charset.get('encoding'):
                content = self.file.read().decode(charset['encoding'])
            else:
                content = self.file.read().decode('utf-8', 'surrogatescape')
        else:
            content = self.file.read()
        return (self.filename(), content, mime_type)

    def filename(self):
        return self.original_name

    def __str__(self):
        return self.filename()


class PreviewMail(BaseMail):
    """
    Like a Mail, but created by a preview_send API call

    We keep them appart not to mess up stats and so. PreviewMail are not even
    exposed through API.
    """
    def as_message(self, *args, **kwargs):
        message = super().as_message(*args, **kwargs)
        message.subject = '[TEST] ' + message.subject
        return message

    def save(self, *args, **kwargs):
        if not self.id:
            # put a clear preview identifier
            self.identifier = 'preview' + mk_base64_uuid()
        super().save(*args, **kwargs)

    @classmethod
    def send_preview(cls, mail_id):
        mail = cls.objects.get(pk=mail_id)
        message = mail.as_message()

        backend = Backend()
        backend.send_simple_message(message)
        log.debug('[{}] Sent preview to {} for message {}'.format(
            mail.identifier, mail.recipient, mail.message.identifier))
        return 'Sent preview {}'.format(mail.identifier)



from slimta.policy import QueuePolicy


class Clean(QueuePolicy):
    """
    Try to determine if there are multiple recipient to clean "To" header
    """
    def apply(self, envelope):
        if envelope.recipients:
            envelope.headers.replace_header('To', envelope.recipients[0])



import yaml
from django.core.management.base import BaseCommand
from django.contrib.auth.models import Group


class Command(BaseCommand):
    help = "List current permissions."

    def add_arguments(self, parser):
        parser.add_argument(
            '--app', nargs='+', type=str, help='Filter app name.')
        parser.add_argument(
            '--group', nargs='+', type=str, help='Filter group name.')
        parser.add_argument(
            '--model', nargs='+', type=str, help='Filter model name.')

    def handle(self, *args, **options):
        result = {}

        permission_kwargs = {}
        if options.get('model', []):
            permission_kwargs[
                'content_type__model__in'] = options.get('model')
        if options.get('app', []):
            permission_kwargs[
                'content_type__app_label__in'] = options.get('app')

        group_kwargs = {}
        if options.get('group', []):
            group_kwargs['name__in'] = options.get('group')

        for group in Group.objects.filter(**group_kwargs):
            result.setdefault(group.name, {})
            for permission in group.permissions.filter(**permission_kwargs):
                content_type = permission.content_type
                result[group.name].setdefault(content_type.app_label, {})
                result[group.name][content_type.app_label].setdefault(
                    content_type.model, [])
                result[group.name][content_type.app_label][
                    content_type.model].append(permission.codename)

        print(yaml.dump(result))
